import glob
import os


os.chdir('bin')

ignored = []

def do_strip(fn):
    try:
        os.system('strip "%s"' % fn)
    except Exception, e:
        print e
    else:
        print 'strip: ', fn

for dll in glob.glob('*.dll'):
    if dll.startswith('python') or dll.startswith('msvc'):
        ignored.append(dll)
        continue
    else:
        do_strip(dll)

for exe in glob.glob('*.exe'):
    do_strip(exe)
    
print '----------------------------'
print 'ignored:'
for i in ignored:
    print '>> ', i



'''

seaf-cli is command line interface for seafile client.

Subcommands:

    init:           create config files for seafile client
    start:          start and run seafile client as daemon
    stop:           stop seafile client
    list:           list local liraries
    status:         show syncing status
    download:       download a library from seafile server
    sync:           synchronize an existing folder with a library in
                        seafile server
    desync:         desynchronize a library with seafile server
    create:         create a new library


Detail
======

Seafile client stores all its configure information in a config dir. The default location is `~/.ccnet`. All the commands below accept an option `-c <config-dir>`.

init
----
Initialize seafile client. This command initializes the config dir. It also creates sub-directories `seafile-data` and `seafile` under `parent-dir`. `seafile-data` is used to store internal data, while `seafile` is used as the default location put downloaded libraries.

    seaf-cli init [-c <config-dir>] -d <parent-dir>

start
-----
Start seafile client. This command start `ccnet` and `seaf-daemon`, `ccnet` is the network part of seafile client, `seaf-daemon` manages the files.

    seaf-cli start [-c <config-dir>]

stop
----
Stop seafile client.

    seaf-cli stop [-c <config-dir>]


Download
--------
Download a library from seafile server

    seaf-cli download -l <library-id> -s <seahub-server-url> -d <parent-directory> -u <username> -p <password>


sync
----
Synchronize a library with an existing folder.

    seaf-cli sync -l <library-id> -s <seahub-server-url> -d <existing-folder> -u <username> -p <password>

desync
------
Desynchronize a library from seafile server

    seaf-cli desync -d <existing-folder>

create
------
Create a new library

    seaf-cli create -s <seahub-server-url> -n <library-name> -u <username> -p <password> -t <description> [-e <library-password>]

'''

import os
import json
import subprocess
import sys
import time
import urllib
import urllib2
import httplib
from urlparse import urlparse

import ccnet
import seafile

def _check_seafile():
    ''' Check ccnet and seafile have been installed '''

    sep = ':' if os.name != 'nt' else ';'
    dirs = os.environ['PATH'].split(sep)
    def exist_in_path(prog):
        ''' Check whether 'prog' exists in system path '''
        for d in dirs:
            if d == '':
                continue
            path = os.path.join(d, prog)
            if os.path.exists(path):
                return True

    progs = [ 'ccnet', 'seaf-daemon' ]

    for prog in progs:
        if os.name == 'nt':
            prog += '.exe'
        if not exist_in_path(prog):
            print "%s not found in PATH. Have you installed seafile?" % prog
            sys.exit(1)

def run_argv(argv, cwd=None, env=None, suppress_stdout=False, suppress_stderr=False, wait=True):
    '''Run a program and wait it to finish, and return its exit code. The
    standard output of this program is supressed.

    '''
    with open(os.devnull, 'w') as devnull:
        if suppress_stdout:
            stdout = devnull
        else:
            stdout = sys.stdout

        if suppress_stderr:
            stderr = devnull
        else:
            stderr = sys.stderr

        proc = subprocess.Popen(argv,
                                cwd=cwd,
                                stdout=stdout,
                                stderr=stderr,
                                env=env)
        if wait:
            return proc.wait()
        return 0

def get_env():
    env = dict(os.environ)
    ld_library_path = os.environ.get('SEAFILE_LD_LIBRARY_PATH', '')
    if ld_library_path:
        env['LD_LIBRARY_PATH'] = ld_library_path

    return env

class NoRedirect(urllib2.HTTPRedirectHandler):
    def redirect_request(self, req, fp, code, msg, hdrs, newurl):
        pass

def urlopen(url, data=None, headers=None, follow_redirect=True):
    if data:
        data = urllib.urlencode(data)
    headers = headers or {}
    req = urllib2.Request(url, data=data, headers=headers)
    if follow_redirect:
        resp = urllib2.urlopen(req)
    else:
        try:
            opener = urllib2.build_opener(NoRedirect())
            resp = opener.open(req)
        except urllib2.HTTPError:
            return None
    return resp.read()

def get_token(url, username, password):
    data = {
        'username': username,
        'password': password,
    }
    token_json = urlopen("%s/api2/auth-token/" % url, data=data)
    tmp = json.loads(token_json)
    token = tmp['token']
    return token

def get_repo_downlod_info(url, token):
    headers = { 'Authorization': 'Token %s' % token }
    repo_info = urlopen(url, headers=headers)
    return json.loads(repo_info)

def seaf_init(conf_dir):
    ''' initialize config directorys'''

    _check_seafile()

    seafile_ini = os.path.join(conf_dir, "seafile.ini")
    seafile_data = os.path.join(conf_dir, "seafile-data")
    fp = open(seafile_ini, 'w')
    fp.write(seafile_data)
    fp.close()

    print 'Init ccnet config success.'

def seaf_start_all(conf_dir):
    ''' start ccnet and seafile daemon '''

    seaf_start_ccnet(conf_dir)
    # wait ccnet process
    time.sleep(1)
    seaf_start_seafile(conf_dir)

    print 'Start ccnet, seafile daemon success.'

def seaf_start_ccnet(conf_dir):
    ''' start ccnet daemon '''

    cmd = [ "ccnet", "--daemon", "-c", conf_dir ]
    wait = False if os.name == 'nt' else True
    if run_argv(cmd, env=get_env(), suppress_stdout=True, wait=wait) != 0:
        print 'Failed to start ccnet daemon.'
        sys.exit(1)

def seaf_start_seafile(conf_dir):
    ''' start seafile daemon '''

    cmd = [ "seaf-daemon", "--daemon", "-c", conf_dir,
            "-d", os.path.join(conf_dir, 'seafile-data'),
            "-w", os.path.join(conf_dir, 'seafile') ]
    wait = False if os.name == 'nt' else True
    if run_argv(cmd, env=get_env(), suppress_stdout=True, wait=wait) != 0:
        print 'Failed to start seafile daemon'
        sys.exit(1)

def seaf_stop(conf_dir):
    '''stop seafile daemon '''

    pool = ccnet.ClientPool(conf_dir)
    client = pool.get_client()
    try:
        client.send_cmd("shutdown")
    except:
        # ignore NetworkError("Failed to read from socket")
        pass

    print 'Stop ccnet, seafile daemon success.'

def get_base_url(url):
    parse_result = urlparse(url)
    scheme = parse_result.scheme
    netloc = parse_result.netloc

    if scheme and netloc:
        return '%s://%s' % (scheme, netloc)

    return None

def get_netloc(url):
    parse_result = urlparse(url)
    return parse_result.netloc

def seaf_sync(conf_dir, server_url, repo_id, worktree, username, passwd):
    ''' synchronize a library from seafile server '''

    pool = ccnet.ClientPool(conf_dir)
    seafile_rpc = seafile.RpcClient(pool, req_pool=False)

    token = get_token(server_url, username, passwd)
    tmp = get_repo_downlod_info("%s/api2/repos/%s/download-info/" % (server_url, repo_id), token)

    encrypted = tmp['encrypted']
    magic = tmp.get('magic', None)
    enc_version = tmp.get('enc_version', None)
    random_key = tmp.get('random_key', None)

    clone_token = tmp['token']
    relay_id = tmp['relay_id']
    relay_addr = tmp['relay_addr']
    relay_port = str(tmp['relay_port'])
    email = tmp['email']
    repo_name = tmp['repo_name']
    version = tmp.get('repo_version', 0)

    more_info = None
    base_url = get_base_url(server_url)
    if base_url:
        more_info = json.dumps({'server_url': base_url})

    if encrypted == 1:
        repo_passwd = 's123'
    else:
        repo_passwd = None

    seafile_rpc.clone(repo_id,
                      version,
                      relay_id,
                      repo_name.encode('utf-8'),
                      worktree,
                      clone_token,
                      repo_passwd, magic,
                      relay_addr,
                      relay_port,
                      email, random_key, enc_version, more_info)

    print 'Synchronize repo test success.'

def seaf_desync(conf_dir, repo_path):
    '''Desynchronize a library from seafile server'''

    pool = ccnet.ClientPool(conf_dir)
    seafile_rpc = seafile.RpcClient(pool, req_pool=False)

    repos = seafile_rpc.get_repo_list(-1, -1)
    repo = None
    for r in repos:
        if r.worktree.replace('/', '\\') == repo_path.decode('utf-8').replace('/', '\\'):
            repo = r
            break

    if repo:
        print "Desynchronize repo test success."
        seafile_rpc.remove_repo(repo.id)
    else:
        print "%s is not a library worktree" % repo_path

def seaf_create(conf_dir, server_url, username, passwd, enc_repo):
    '''Create a library'''

    # curl -d 'username=<USERNAME>&password=<PASSWORD>' http://127.0.0.1:8000/api2/auth-token
    token = get_token(server_url, username, passwd)

    headers = { 'Authorization': 'Token %s' % token }
    data = {
        'name': 'test',
        'desc': 'test',
    }
    if enc_repo:
        data['passwd'] = 's123'

    repo_info_json =  urlopen("%s/api2/repos/" % server_url, data=data, headers=headers)
    repo_info = json.loads(repo_info_json)

    if enc_repo:
        print 'Create encrypted repo test success.'
    else:
        print 'Create non encrypted repo test success.'

    return repo_info['repo_id']

def seaf_delete(conf_dir, server_url, username, passwd, repo_id):
    '''Delete a library'''

    token = get_token(server_url, username, passwd)
    headers = { 'Authorization': 'Token %s' % token }

    conn = httplib.HTTPConnection(get_netloc(server_url))
    conn.request('DELETE', '/api2/repos/%s/' % repo_id, None, headers)
    resp = conn.getresponse()
    if resp.status == 200:
        print 'Delete repo test success.'
    else:
        print 'Delete repo test failed: %s.' % resp.reason

def seaf_get_repo(conf_dir, repo_id):
    pool = ccnet.ClientPool(conf_dir)
    seafile_rpc = seafile.RpcClient(pool, req_pool=False)
    return seafile_rpc.seafile_get_repo(repo_id)



#!/usr/bin/env python
# coding: UTF-8

'''This scirpt builds the seafile windows msi installer.

Some notes:

1. The working directory is always the 'builddir'. 'os.chdir' is only called
to change to the 'builddir'. We make use of the 'cwd' argument in
'subprocess.Popen' to run a command in a specific directory.

2. When invoking commands like 'tar', we must convert the path to posix path with the function to_mingw_path. E.g., 'c:\\seafile' should be converted to '/c/seafile'.

'''

import sys

####################
### Requires Python 2.6+
####################
if sys.version_info[0] == 3:
    print 'Python 3 not supported yet. Quit now.'
    sys.exit(1)
if sys.version_info[1] < 6:
    print 'Python 2.6 or above is required. Quit now.'
    sys.exit(1)

import multiprocessing
import os
import glob
import shutil
import re
import subprocess
import optparse
import atexit
import csv
import time

error_exit = False
####################
### Global variables
####################

# command line configuartion
conf = {}

# The retry times when sign programs
RETRY_COUNT = 3

# key names in the conf dictionary.
CONF_VERSION            = 'version'
CONF_LIBSEARPC_VERSION  = 'libsearpc_version'
CONF_SEAFILE_VERSION    = 'seafile_version'
CONF_SEAFILE_CLIENT_VERSION  = 'seafile_client_version'
CONF_SRCDIR             = 'srcdir'
CONF_KEEP               = 'keep'
CONF_BUILDDIR           = 'builddir'
CONF_OUTPUTDIR          = 'outputdir'
CONF_DEBUG              = 'debug'
CONF_ONLY_CHINESE       = 'onlychinese'
CONF_QT_ROOT            = 'qt_root'
CONF_EXTRA_LIBS_DIR     = 'extra_libs_dir'
CONF_QT5                = 'qt5'
CONF_BRAND              = 'brand'
CONF_CERTFILE           = 'certfile'
CONF_NO_STRIP           = 'nostrip'

####################
### Common helper functions
####################
def to_mingw_path(path):
    if len(path) < 2 or path[1] != ':' :
        return path.replace('\\', '/')

    drive = path[0]
    return '/%s%s' % (drive.lower(), path[2:].replace('\\', '/'))

def to_win_path(path):
    if len(path) < 2 or path[1] == ':' :
        return path.replace('/', '\\')

    drive = path[1]
    return '%s:%s' % (drive.lower(), path[2:].replace('/', '\\'))

def highlight(content, is_error=False):
    '''Add ANSI color to content to get it highlighted on terminal'''
    dummy = is_error
    return content
    # if is_error:
    #     return '\x1b[1;31m%s\x1b[m' % content
    # else:
    #     return '\x1b[1;32m%s\x1b[m' % content

def info(msg):
    print highlight('[INFO] ') + msg

def find_in_path(prog):
    '''Test whether prog exists in system path'''
    dirs = os.environ['PATH'].split(';')
    for d in dirs:
        if d == '':
            continue
        path = os.path.join(d, prog)
        if os.path.exists(path):
            return path

    return None

def prepend_env_value(name, value, seperator=':'):
    '''prepend a new value to a list'''
    try:
        current_value = os.environ[name]
    except KeyError:
        current_value = ''

    new_value = value
    if current_value:
        new_value += seperator + current_value

    os.environ[name] = new_value

def error(msg=None, usage=None):
    if msg:
        print highlight('[ERROR] ') + msg
    if usage:
        print usage
    sys.exit(1)

def run_argv(argv, cwd=None, env=None, suppress_stdout=False, suppress_stderr=False):
    '''Run a program and wait it to finish, and return its exit code. The
    standard output of this program is supressed.

    '''
    with open(os.devnull, 'w') as devnull:
        if suppress_stdout:
            stdout = devnull
        else:
            stdout = sys.stdout

        if suppress_stderr:
            stderr = devnull
        else:
            stderr = sys.stderr

        proc = subprocess.Popen(argv,
                                cwd=cwd,
                                stdout=stdout,
                                stderr=stderr,
                                env=env)
        return proc.wait()

def run(cmdline, cwd=None, env=None, suppress_stdout=False, suppress_stderr=False):
    '''Like run_argv but specify a command line string instead of argv'''
    info('running %s, cwd=%s' % (cmdline, cwd if cwd else os.getcwd()))
    with open(os.devnull, 'w') as devnull:
        if suppress_stdout:
            stdout = devnull
        else:
            stdout = sys.stdout

        if suppress_stderr:
            stderr = devnull
        else:
            stderr = sys.stderr

        proc = subprocess.Popen(cmdline,
                                cwd=cwd,
                                stdout=stdout,
                                stderr=stderr,
                                env=env,
                                shell=True)
        ret = proc.wait()
        if 'depend' not in cmdline and ret != 0:
            global error_exit
            error_exit = True
        return ret

def must_mkdir(path):
    '''Create a directory, exit on failure'''
    try:
        os.makedirs(path)
    except OSError, e:
        error('failed to create directory %s:%s' % (path, e))

def must_copy(src, dst):
    '''Copy src to dst, exit on failure'''
    try:
        shutil.copy(src, dst)
    except Exception, e:
        error('failed to copy %s to %s: %s' % (src, dst, e))

def must_copytree(src, dst):
    '''Copy dir src to dst, exit on failure'''
    try:
        shutil.copytree(src, dst)
    except Exception, e:
        error('failed to copy dir %s to %s: %s' % (src, dst, e))

def must_move(src, dst):
    '''Move src to dst, exit on failure'''
    try:
        shutil.move(src, dst)
    except Exception, e:
        error('failed to move %s to %s: %s' % (src, dst, e))

class Project(object):
    '''Base class for a project'''
    # Probject name, i.e. libseaprc/seafile/seahub
    name = ''

    # A list of shell commands to configure/build the project
    build_commands = []

    def __init__(self):
        self.prefix = os.path.join(conf[CONF_BUILDDIR], 'usr')
        self.version = self.get_version()
        self.src_tarball = os.path.join(conf[CONF_SRCDIR],
                            '%s-%s.tar.gz' % (self.name, self.version))

        # project dir, like <builddir>/seafile-1.2.2/
        self.projdir = os.path.join(conf[CONF_BUILDDIR], '%s-%s' % (self.name, self.version))

    def get_version(self):
        # libsearpc can have different versions from seafile.
        raise NotImplementedError

    def get_source_commit_id(self):
        '''By convetion, we record the commit id of the source code in the
        file "<projdir>/latest_commit"

        '''
        latest_commit_file = os.path.join(self.projdir, 'latest_commit')
        with open(latest_commit_file, 'r') as fp:
            commit_id = fp.read().strip('\n\r\t ')

        return commit_id

    def append_cflags(self, macros):
        cflags = ' '.join([ '-D%s=%s' % (k, macros[k]) for k in macros ])
        prepend_env_value('CPPFLAGS',
                          cflags,
                          seperator=' ')

    def uncompress(self):
        '''Uncompress the source from the tarball'''
        info('Uncompressing %s' % self.name)

        tarball = to_mingw_path(self.src_tarball)
        if run('tar xf %s' % tarball) != 0:
            error('failed to uncompress source of %s' % self.name)

    def before_build(self):
        '''Hook method to do project-specific stuff before running build commands'''
        pass

    def build(self):
        '''Build the source'''
        self.before_build()
        info('Building %s' % self.name)
        dump_env()
        for cmd in self.build_commands:
            if run(cmd, cwd=self.projdir) != 0:
                error('error when running command:\n\t%s\n' % cmd)

def get_make_path():
    return find_in_path('make.exe')

def concurrent_make():
    return '%s -j%s' % (get_make_path(), multiprocessing.cpu_count())

class Libsearpc(Project):
    name = 'libsearpc'

    def __init__(self):
        Project.__init__(self)
        self.build_commands = [
            'sh ./configure --prefix=%s --disable-compile-demo' % to_mingw_path(self.prefix),
            concurrent_make(),
            '%s install' % get_make_path(),
        ]

    def get_version(self):
        return conf[CONF_LIBSEARPC_VERSION]

class Seafile(Project):
    name = 'seafile'
    def __init__(self):
        Project.__init__(self)
        enable_breakpad = '--enable-breakpad'
        self.build_commands = [
            'sh ./configure %s --prefix=%s' % (enable_breakpad, to_mingw_path(self.prefix)),
            concurrent_make(),
            '%s install' % get_make_path(),
        ]

    def get_version(self):
        return conf[CONF_SEAFILE_VERSION]

    def before_build(self):
        macros = {}
        # SET SEAFILE_SOURCE_COMMIT_ID, so it can be printed in the log
        macros['SEAFILE_SOURCE_COMMIT_ID'] = '\\"%s\\"' % self.get_source_commit_id()
        self.append_cflags(macros)

class SeafileClient(Project):
    name = 'seafile-client'
    def __init__(self):
        Project.__init__(self)
        ninja = find_in_path('ninja.exe')
        seafile_prefix = Seafile().prefix
        generator = 'Ninja' if ninja else 'MSYS Makefiles'
        build_type = 'Debug' if conf[CONF_DEBUG] else 'Release'
        flags = {
            'BUILD_SPARKLE_SUPPORT': 'ON',
            'USE_QT5': 'ON' if conf[CONF_QT5] else 'OFF',
            'BUILD_SHIBBOLETH_SUPPORT': 'ON',
            'CMAKE_BUILD_TYPE': build_type,
            'CMAKE_INSTALL_PREFIX': to_mingw_path(self.prefix),
            # ninja invokes cmd.exe which doesn't support msys/mingw path
            # change the value but don't override CMAKE_EXE_LINKER_FLAGS,
            # which is in use
            'CMAKE_EXE_LINKER_FLAGS_%s' % build_type.upper(): '-L%s' % (os.path.join(seafile_prefix, 'lib') if ninja else to_mingw_path(os.path.join(seafile_prefix, 'lib'))),
        }
        flags_str = ' '.join(['-D%s=%s' % (k, v) for k, v in flags.iteritems()])
        make = ninja or concurrent_make()
        self.build_commands = [
            'cmake -G "%s" %s .' % (generator, flags_str),
            make,
            '%s install' % make,
            "bash extensions/build.sh",
        ]

    def get_version(self):
        return conf[CONF_SEAFILE_CLIENT_VERSION]

    def before_build(self):
        shutil.copy(os.path.join(conf[CONF_EXTRA_LIBS_DIR], 'winsparkle.lib'), self.projdir)

class SeafileShellExt(Project):
    name = 'seafile-shell-ext'
    def __init__(self):
        Project.__init__(self)
        self.build_commands = [
            "bash extensions/build.sh",
            "bash shellext-fix/build.sh",
        ]

    def get_version(self):
        return conf[CONF_SEAFILE_CLIENT_VERSION]

def check_targz_src(proj, version, srcdir):
    src_tarball = os.path.join(srcdir, '%s-%s.tar.gz' % (proj, version))
    if not os.path.exists(src_tarball):
        error('%s not exists' % src_tarball)

def validate_args(usage, options):
    required_args = [
        CONF_VERSION,
        CONF_LIBSEARPC_VERSION,
        CONF_SEAFILE_VERSION,
        CONF_SEAFILE_CLIENT_VERSION,
        CONF_SRCDIR,
        CONF_QT_ROOT,
        CONF_EXTRA_LIBS_DIR,
    ]

    # fist check required args
    for optname in required_args:
        if getattr(options, optname, None) == None:
            error('%s must be specified' % optname, usage=usage)

    def get_option(optname):
        return getattr(options, optname)

    # [ version ]
    def check_project_version(version):
        '''A valid version must be like 1.2.2, 1.3'''
        if not re.match(r'^[0-9]+(\.[0-9]+)+$', version):
            error('%s is not a valid version' % version, usage=usage)

    version = get_option(CONF_VERSION)
    libsearpc_version = get_option(CONF_LIBSEARPC_VERSION)
    seafile_version = get_option(CONF_SEAFILE_VERSION)
    seafile_client_version = get_option(CONF_SEAFILE_CLIENT_VERSION)
    seafile_shell_ext_version = get_option(CONF_SEAFILE_CLIENT_VERSION)

    check_project_version(version)
    check_project_version(libsearpc_version)
    check_project_version(seafile_version)
    check_project_version(seafile_client_version)
    check_project_version(seafile_shell_ext_version)

    # [ srcdir ]
    srcdir = to_win_path(get_option(CONF_SRCDIR))
    check_targz_src('libsearpc', libsearpc_version, srcdir)
    check_targz_src('seafile', seafile_version, srcdir)
    check_targz_src('seafile-client', seafile_client_version, srcdir)
    check_targz_src('seafile-shell-ext', seafile_shell_ext_version, srcdir)

    # [ builddir ]
    builddir = to_win_path(get_option(CONF_BUILDDIR))
    if not os.path.exists(builddir):
        error('%s does not exist' % builddir, usage=usage)

    builddir = os.path.join(builddir, 'seafile-msi-build')

    # [ outputdir ]
    outputdir = to_win_path(get_option(CONF_OUTPUTDIR))
    if not os.path.exists(outputdir):
        error('outputdir %s does not exist' % outputdir, usage=usage)

    # [ keep ]
    keep = get_option(CONF_KEEP)

    # [ no strip]
    debug = get_option(CONF_DEBUG)

    # [ no strip]
    nostrip = get_option(CONF_NO_STRIP)

    # [only chinese]
    onlychinese = get_option(CONF_ONLY_CHINESE)

    # [ qt root]
    qt_root = get_option(CONF_QT_ROOT)
    def check_qt_root(qt_root):
        if not os.path.exists(os.path.join(qt_root, 'plugins')):
            error('%s is not a valid qt root' % qt_root)
    check_qt_root(qt_root)

    # [ sparkle dir]
    extra_libs_dir = get_option(CONF_EXTRA_LIBS_DIR)
    def check_extra_libs_dir(extra_libs_dir):
        for fn in ['winsparkle.lib']:
            if not os.path.exists(os.path.join(extra_libs_dir, fn)):
                error('%s is missing in %s' % (fn, extra_libs_dir))
    check_extra_libs_dir(extra_libs_dir)

    # [qt5]
    qt5 = get_option(CONF_QT5)
    brand = get_option(CONF_BRAND)
    cert = get_option(CONF_CERTFILE)
    if cert is not None:
        if not os.path.exists(cert):
            error('cert file "{}" does not exist'.format(cert))

    conf[CONF_VERSION] = version
    conf[CONF_LIBSEARPC_VERSION] = libsearpc_version
    conf[CONF_SEAFILE_VERSION] = seafile_version
    conf[CONF_SEAFILE_CLIENT_VERSION] = seafile_client_version

    conf[CONF_BUILDDIR] = builddir
    conf[CONF_SRCDIR] = srcdir
    conf[CONF_OUTPUTDIR] = outputdir
    conf[CONF_KEEP] = True
    conf[CONF_DEBUG] = debug or nostrip
    conf[CONF_NO_STRIP] = debug or nostrip
    conf[CONF_ONLY_CHINESE] = onlychinese
    conf[CONF_QT_ROOT] = qt_root
    conf[CONF_EXTRA_LIBS_DIR] = extra_libs_dir
    conf[CONF_QT5] = qt5
    conf[CONF_BRAND] = brand
    conf[CONF_CERTFILE] = cert

    prepare_builddir(builddir)
    show_build_info()

def show_build_info():
    '''Print all conf information. Confirm before continue.'''
    info('------------------------------------------')
    info('Seafile msi installer: BUILD INFO')
    info('------------------------------------------')
    info('seafile:                  %s' % conf[CONF_VERSION])
    info('libsearpc:                %s' % conf[CONF_LIBSEARPC_VERSION])
    info('seafile:                  %s' % conf[CONF_SEAFILE_VERSION])
    info('seafile-client:           %s' % conf[CONF_SEAFILE_CLIENT_VERSION])
    info('qt-root:                  %s' % conf[CONF_QT_ROOT])
    info('builddir:                 %s' % conf[CONF_BUILDDIR])
    info('outputdir:                %s' % conf[CONF_OUTPUTDIR])
    info('source dir:               %s' % conf[CONF_SRCDIR])
    info('debug:                    %s' % conf[CONF_DEBUG])
    info('build english version:    %s' % (not conf[CONF_ONLY_CHINESE]))
    info('clean on exit:            %s' % (not conf[CONF_KEEP]))
    info('------------------------------------------')
    info('press any key to continue ')
    info('------------------------------------------')
    dummy = raw_input()

def prepare_builddir(builddir):
    must_mkdir(builddir)

    if not conf[CONF_KEEP]:
        def remove_builddir():
            '''Remove the builddir when exit'''
            if not error_exit:
                info('remove builddir before exit')
                shutil.rmtree(builddir, ignore_errors=True)
        atexit.register(remove_builddir)

    os.chdir(builddir)

def parse_args():
    parser = optparse.OptionParser()
    def long_opt(opt):
        return '--' + opt

    parser.add_option(long_opt(CONF_VERSION),
                      dest=CONF_VERSION,
                      nargs=1,
                      help='the version to build. Must be digits delimited by dots, like 1.3.0')

    parser.add_option(long_opt(CONF_LIBSEARPC_VERSION),
                      dest=CONF_LIBSEARPC_VERSION,
                      nargs=1,
                      help='the version of libsearpc as specified in its "configure.ac". Must be digits delimited by dots, like 1.3.0')

    parser.add_option(long_opt(CONF_SEAFILE_VERSION),
                      dest=CONF_SEAFILE_VERSION,
                      nargs=1,
                      help='the version of seafile as specified in its "configure.ac". Must be digits delimited by dots, like 1.3.0')

    parser.add_option(long_opt(CONF_SEAFILE_CLIENT_VERSION),
                      dest=CONF_SEAFILE_CLIENT_VERSION,
                      nargs=1,
                      help='the version of seafile-client. Must be digits delimited by dots, like 1.3.0')

    parser.add_option(long_opt(CONF_BUILDDIR),
                      dest=CONF_BUILDDIR,
                      nargs=1,
                      help='the directory to build the source. Defaults to /c',
                      default='c:\\')

    parser.add_option(long_opt(CONF_OUTPUTDIR),
                      dest=CONF_OUTPUTDIR,
                      nargs=1,
                      help='the output directory to put the generated server tarball. Defaults to the current directory.',
                      default=os.getcwd())

    parser.add_option(long_opt(CONF_SRCDIR),
                      dest=CONF_SRCDIR,
                      nargs=1,
                      help='''Source tarballs must be placed in this directory.''')

    parser.add_option(long_opt(CONF_QT_ROOT),
                      dest=CONF_QT_ROOT,
                      nargs=1,
                      help='''qt root directory.''')

    parser.add_option(long_opt(CONF_EXTRA_LIBS_DIR),
                      dest=CONF_EXTRA_LIBS_DIR,
                      nargs=1,
                      help='''where we can find winsparkle.lib''')

    parser.add_option(long_opt(CONF_KEEP),
                      dest=CONF_KEEP,
                      action='store_true',
                      help='''keep the build directory after the script exits. By default, the script would delete the build directory at exit.''')

    parser.add_option(long_opt(CONF_DEBUG),
                      dest=CONF_DEBUG,
                      action='store_true',
                      help='''compile in debug mode''')

    parser.add_option(long_opt(CONF_ONLY_CHINESE),
                      dest=CONF_ONLY_CHINESE,
                      action='store_true',
                      help='''only build the Chinese version. By default both Chinese and English versions would be built.''')

    parser.add_option(long_opt(CONF_QT5),
                      dest=CONF_QT5,
                      action='store_true',
                      help='''build seafile client with qt5''')

    parser.add_option(long_opt(CONF_BRAND),
                      dest=CONF_BRAND,
                      default='seafile',
                      help='''brand name of the package''')

    parser.add_option(long_opt(CONF_CERTFILE),
                      nargs=1,
                      default=None,
                      dest=CONF_CERTFILE,
                      help='''The cert for signing the executables and the installer.''')

    parser.add_option(long_opt(CONF_NO_STRIP),
                      dest=CONF_NO_STRIP,
                      action='store_true',
                      help='''do not strip the symbols.''')

    usage = parser.format_help()
    options, remain = parser.parse_args()
    if remain:
        error(usage=usage)

    validate_args(usage, options)

def setup_build_env():
    '''Setup environment variables, such as export PATH=$BUILDDDIR/bin:$PATH'''
    prefix = Seafile().prefix
    prepend_env_value('CPPFLAGS',
                     '-I%s' % to_mingw_path(os.path.join(prefix, 'include')),
                     seperator=' ')

    prepend_env_value('CPPFLAGS',
                     '-DSEAFILE_CLIENT_VERSION=\\"%s\\"' % conf[CONF_VERSION],
                     seperator=' ')

    prepend_env_value('CPPFLAGS',
                      '-g -fno-omit-frame-pointer',
                      seperator=' ')
    if conf[CONF_DEBUG]:
        prepend_env_value('CPPFLAGS', '-O0', seperator=' ')

    prepend_env_value('LDFLAGS',
                     '-L%s' % to_mingw_path(os.path.join(prefix, 'lib')),
                     seperator=' ')

    prepend_env_value('PATH',
                      os.path.join(prefix, 'bin'),
                      seperator=';')

    prepend_env_value('PKG_CONFIG_PATH',
                      os.path.join(prefix, 'lib', 'pkgconfig'),
                      seperator=';')
                      # to_mingw_path(os.path.join(prefix, 'lib', 'pkgconfig')))

    # specifiy the directory for wix temporary files
    wix_temp_dir = os.path.join(conf[CONF_BUILDDIR], 'wix-temp')
    os.environ['WIX_TEMP'] = wix_temp_dir

    must_mkdir(wix_temp_dir)

def dependency_walk(applet):
    output = os.path.join(conf[CONF_BUILDDIR], 'depends.csv')
    cmd = 'depends.exe -c -f 1 -oc %s %s' % (output, applet)

    # See the manual of Dependency walker
    if run(cmd) > 0x100:
        error('failed to run dependency walker for %s' % applet)

    if not os.path.exists(output):
        error('failed to run dependency walker for %s' % applet)

    shared_libs = parse_depends_csv(output)
    return shared_libs

def parse_depends_csv(path):
    '''parse the output of dependency walker'''
    libs = set()
    our_libs = ['libsearpc', 'libseafile']
    def should_ignore_lib(lib):
        lib = lib.lower()
        if not os.path.exists(lib):
            return True

        if lib.startswith('c:\\windows'):
            return True

        if lib.endswith('exe'):
            return True

        for name in our_libs:
            if name in lib:
                return True

        return False

    with open(path, 'r') as fp:
        reader = csv.reader(fp)
        for row in reader:
            if len(row) < 2:
                continue
            lib = row[1]
            if not should_ignore_lib(lib):
                libs.add(lib)

    return set(libs)

def copy_shared_libs(exes):
    '''Copy shared libs need by seafile-applet.exe, such as libsearpc,
    libseafile, etc. First we use Dependency walker to analyse
    seafile-applet.exe, and get an output file in csv format. Then we parse
    the csv file to get the list of shared libs.

    '''

    shared_libs = set()
    for exectuable in exes:
        shared_libs.update(dependency_walk(exectuable))

    pack_bin_dir = os.path.join(conf[CONF_BUILDDIR], 'pack', 'bin')
    for lib in shared_libs:
        must_copy(lib, pack_bin_dir)

    if not any([os.path.basename(lib).lower().startswith('libssl') for lib in shared_libs]):
        ssleay32 = find_in_path('ssleay32.dll')
        must_copy(ssleay32, pack_bin_dir)

def copy_dll_exe():
    prefix = Seafile().prefix
    destdir = os.path.join(conf[CONF_BUILDDIR], 'pack', 'bin')

    filelist = [
        os.path.join(prefix, 'bin', 'libsearpc-1.dll'),
        os.path.join(prefix, 'bin', 'libseafile-0.dll'),
        os.path.join(prefix, 'bin', 'seaf-daemon.exe'),
        os.path.join(SeafileClient().projdir, 'seafile-applet.exe'),
        os.path.join(SeafileShellExt().projdir, 'shellext-fix', 'shellext-fix.exe'),
    ]

    for name in filelist:
        must_copy(name, destdir)

    extdlls = [
        os.path.join(SeafileShellExt().projdir, 'extensions', 'lib', 'seafile_ext.dll'),
        os.path.join(SeafileShellExt().projdir, 'extensions', 'lib', 'seafile_ext64.dll'),
    ]

    customdir = os.path.join(conf[CONF_BUILDDIR], 'pack', 'custom')
    for dll in extdlls:
        must_copy(dll, customdir)

    copy_shared_libs([ f for f in filelist if f.endswith('.exe') ])
    copy_qt_plugins_imageformats()
    copy_qt_plugins_platforms()
    copy_qt_translations()

def copy_qt_plugins_imageformats():
    destdir = os.path.join(conf[CONF_BUILDDIR], 'pack', 'bin', 'imageformats')
    must_mkdir(destdir)

    qt_plugins_srcdir = os.path.join(conf[CONF_QT_ROOT], 'plugins', 'imageformats')

    src = os.path.join(qt_plugins_srcdir, 'qico4.dll')
    if conf[CONF_QT5]:
        src = os.path.join(qt_plugins_srcdir, 'qico.dll')
    must_copy(src, destdir)

    src = os.path.join(qt_plugins_srcdir, 'qgif4.dll')
    if conf[CONF_QT5]:
        src = os.path.join(qt_plugins_srcdir, 'qgif.dll')
    must_copy(src, destdir)

    src = os.path.join(qt_plugins_srcdir, 'qjpeg.dll')
    if conf[CONF_QT5]:
        src = os.path.join(qt_plugins_srcdir, 'qjpeg.dll')
    must_copy(src, destdir)

def copy_qt_plugins_platforms():
    if not conf[CONF_QT5]:
        return

    destdir = os.path.join(conf[CONF_BUILDDIR], 'pack', 'bin', 'platforms')
    must_mkdir(destdir)

    qt_plugins_srcdir = os.path.join(conf[CONF_QT_ROOT], 'plugins', 'platforms')

    src = os.path.join(qt_plugins_srcdir, 'qwindows.dll')
    must_copy(src, destdir)

    src = os.path.join(qt_plugins_srcdir, 'qminimal.dll')
    must_copy(src, destdir)

def copy_qt_translations():
    destdir = os.path.join(conf[CONF_BUILDDIR], 'pack', 'bin')

    qt_translation_dir = os.path.join(conf[CONF_QT_ROOT], 'translations')

    i18n_dir = os.path.join(SeafileClient().projdir, 'i18n')
    qm_pattern = os.path.join(i18n_dir, 'seafile_*.qm')

    qt_qms = set()
    def add_lang(lang):
        if not lang:
            return
        qt_qm = os.path.join(qt_translation_dir, 'qt_%s.qm' % lang)
        if os.path.exists(qt_qm):
            qt_qms.add(qt_qm)
        elif '_' in lang:
            add_lang(lang[:lang.index('_')])

    for fn in glob.glob(qm_pattern):
        name = os.path.basename(fn)
        m = re.match(r'seafile_(.*)\.qm', name)
        lang = m.group(1)
        add_lang(lang)

    for src in qt_qms:
        must_copy(src, destdir)

def prepare_msi():
    pack_dir = os.path.join(conf[CONF_BUILDDIR], 'pack')

    msi_dir = os.path.join(Seafile().projdir, 'msi')

    # These files are in seafile-shell-ext because they're shared between seafile/seadrive
    ext_wxi = os.path.join(SeafileShellExt().projdir, 'msi', 'ext.wxi')
    must_copy(ext_wxi, msi_dir)
    shell_wxs = os.path.join(SeafileShellExt().projdir, 'msi', 'shell.wxs')
    must_copy(shell_wxs, msi_dir)

    must_copytree(msi_dir, pack_dir)
    must_mkdir(os.path.join(pack_dir, 'bin'))

    if run('make', cwd=os.path.join(pack_dir, 'custom')) != 0:
        error('Error when compiling seafile msi custom dlls')

    copy_dll_exe()

def sign_executables():
    certfile = conf.get(CONF_CERTFILE)
    if certfile is None:
        info('exectuable signing is skipped since no cert is provided.')
        return

    pack_dir = os.path.join(conf[CONF_BUILDDIR], 'pack')
    exectuables = glob.glob(os.path.join(pack_dir, 'bin', '*.exe'))
    for exe in exectuables:
        do_sign(certfile, exe)

def sign_installers():
    certfile = conf.get(CONF_CERTFILE)
    if certfile is None:
        info('msi signing is skipped since no cert is provided.')
        return

    pack_dir = os.path.join(conf[CONF_BUILDDIR], 'pack')
    installers = glob.glob(os.path.join(pack_dir, '*.msi'))
    for fn in installers:
        name = conf[CONF_BRAND]
        if name == 'seafile':
            name = 'Seafile'
        do_sign(certfile, fn, desc='{} Installer'.format(name))

def do_sign(certfile, fn, desc=None):
    certfile = to_win_path(certfile)
    fn = to_win_path(fn)
    info('signing file {} using cert "{}"'.format(fn, certfile))

    if desc:
        desc_flags = '-d "{}"'.format(desc)
    else:
        desc_flags = ''

    # https://support.comodo.com/index.php?/Knowledgebase/Article/View/68/0/time-stamping-server
    signcmd = 'signtool.exe sign -fd sha256 -t http://timestamp.comodoca.com -f {} {} {}'.format(certfile, desc_flags, fn)
    i = 0
    while i < RETRY_COUNT:
        time.sleep(30)
        ret = run(signcmd, cwd=os.path.dirname(fn))
        if ret == 0:
            break
        i = i + 1
        if i == RETRY_COUNT:
            error('Failed to sign file "{}"'.format(fn))

def strip_symbols():
    bin_dir = os.path.join(conf[CONF_BUILDDIR], 'pack', 'bin')
    def do_strip(fn, stripcmd='strip'):
        run('%s "%s"' % (stripcmd, fn))
        info('stripping: %s' % fn)

    for dll in glob.glob(os.path.join(bin_dir, '*.dll')):
        name = os.path.basename(dll).lower()
        if 'qt' in name:
            do_strip(dll)
        if name == 'seafile_ext.dll':
            do_strip(dll)
        elif name == 'seafile_ext64.dll':
            do_strip(dll, stripcmd='x86_64-w64-mingw32-strip')

def edit_fragment_wxs():
    '''In the main wxs file(seafile.wxs) we need to reference to the id of
    seafile-applet.exe, which is listed in fragment.wxs. Since fragments.wxs is
    auto generated, the id is sequentially generated, so we need to change the
    id of seafile-applet.exe manually.

    '''
    file_path = os.path.join(conf[CONF_BUILDDIR], 'pack', 'fragment.wxs')
    new_lines = []
    with open(file_path, 'r') as fp:
        for line in fp:
            if 'seafile-applet.exe' in line:
                # change the id of 'seafile-applet.exe' to 'seafileapplet.exe'
                new_line = re.sub('file_bin_[\d]+', 'seafileapplet.exe', line)
                new_lines.append(new_line)
            else:
                new_lines.append(line)

    content = '\r\n'.join(new_lines)
    with open(file_path, 'w') as fp:
        fp.write(content)


def generate_breakpad_symbols():
    """
    Generate seafile and seafile-gui breakpad symbols
    :return: None
    """
    seafile_src = Seafile().projdir
    seafile_gui_src = SeafileClient().projdir
    generate_breakpad_symbols_script = os.path.join(seafile_src, 'scripts/breakpad.py')

    # generate seafile the breakpad symbols
    seafile_name = 'seaf-daemon.exe'
    seafile_symbol_name = 'seaf-daemon.exe.sym-%s' % conf[CONF_VERSION]
    seafile_symbol_output = os.path.join(seafile_src, seafile_symbol_name)

    if run('python %s  --projectSrc %s --name %s --output %s'
           % (generate_breakpad_symbols_script, seafile_src, seafile_name, seafile_symbol_output)) != 0:
        error('Error when generating breakpad symbols')

    # generate seafile gui breakpad symbols
    seafile_gui_name = 'seafile-applet.exe'
    seafile_gui_symbol_name = 'seafile-applet.exe.sym-%s' % conf[CONF_VERSION]
    seafile_gui_symbol_output = os.path.join(seafile_gui_src, seafile_gui_symbol_name)

    if run('python %s --projectSrc %s --name %s --output %s'
            % (generate_breakpad_symbols_script, seafile_gui_src, seafile_gui_name, seafile_gui_symbol_output)) != 0:
        error('Error when generating seafile gui client breakpad symbol')

    # move symbols to output directory
    dst_seafile_symbol_file = os.path.join(conf[CONF_OUTPUTDIR], seafile_symbol_name)
    dst_seafile_gui_symbol_file = os.path.join(conf[CONF_OUTPUTDIR], seafile_gui_symbol_name)
    must_copy(seafile_symbol_output, dst_seafile_symbol_file)
    must_copy(seafile_gui_symbol_output, dst_seafile_gui_symbol_file)


def build_msi():
    prepare_msi()
    generate_breakpad_symbols()
    if conf[CONF_DEBUG] or conf[CONF_NO_STRIP]:
        info('Would not strip exe/dll symbols since --debug or --nostrip is specified')
    else:
        strip_symbols()

    # Only sign the exectuables after stripping symbols.
    if need_sign():
        sign_executables()

    pack_dir = os.path.join(conf[CONF_BUILDDIR], 'pack')
    if run('make fragment.wxs', cwd=pack_dir) != 0:
        error('Error when make fragement.wxs')

    edit_fragment_wxs()

    if run('make', cwd=pack_dir) != 0:
        error('Error when make seafile.msi')

def build_english_msi():
    '''The extra work to build the English msi.'''
    pack_dir = os.path.join(conf[CONF_BUILDDIR], 'pack')

    if run('make en', cwd=pack_dir) != 0:
        error('Error when make seafile-en.msi')

def build_german_msi():
    '''The extra work to build the German msi.'''
    pack_dir = os.path.join(conf[CONF_BUILDDIR], 'pack')

    if run('make de', cwd=pack_dir) != 0:
        error('Error when make seafile-de.msi')

def move_msi():
    pack_dir = os.path.join(conf[CONF_BUILDDIR], 'pack')
    src_msi = os.path.join(pack_dir, 'seafile.msi')
    brand = conf[CONF_BRAND]
    dst_msi = os.path.join(conf[CONF_OUTPUTDIR], '%s-%s.msi' % (brand, conf[CONF_VERSION]))

    # move msi to outputdir
    must_copy(src_msi, dst_msi)

    if not conf[CONF_ONLY_CHINESE]:
        src_msi_en = os.path.join(pack_dir, 'seafile-en.msi')
        dst_msi_en = os.path.join(conf[CONF_OUTPUTDIR], '%s-%s-en.msi' % (brand, conf[CONF_VERSION]))
        must_copy(src_msi_en, dst_msi_en)

    print '---------------------------------------------'
    print 'The build is successfully. Output is:'
    print '>>\t%s' % dst_msi
    if not conf[CONF_ONLY_CHINESE]:
        print '>>\t%s' % dst_msi_en
        # print '>>\t%s' % dst_msi_de
    print '---------------------------------------------'

def check_tools():
    tools = [
        'Paraffin',
        'candle',
        'light',
        'depends',
    ]

    for prog in tools:
        if not find_in_path(prog + '.exe'):
            error('%s not found' % prog)

def dump_env():
    print 'Dumping environment variables:'
    for k, v in os.environ.iteritems():
        print '%s: %s' % (k, v)

def need_sign():
    return conf[CONF_BRAND].lower() == 'seafile'

def main():
    dump_env()
    parse_args()
    setup_build_env()
    check_tools()

    libsearpc = Libsearpc()
    seafile = Seafile()
    seafile_client = SeafileClient()
    seafile_shell_ext = SeafileShellExt()

    libsearpc.uncompress()
    libsearpc.build()

    seafile.uncompress()
    seafile.build()

    seafile_client.uncompress()
    seafile_shell_ext.uncompress()

    seafile_client.build()
    seafile_shell_ext.build()

    build_msi()
    if not conf[CONF_ONLY_CHINESE]:
        build_english_msi()
        # build_german_msi()

    if need_sign():
        sign_installers()
    move_msi()

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding: UTF-8

import sys

####################
### Requires Python 3+
####################
if sys.version_info[0] == 2:
    print('Python 2 not be supported, require Python 3. Quit now.')
    sys.exit(1)

import os
import subprocess
import shutil
import time
import glob
import re

BUILDDIR = os.path.join(os.getcwd(), "..\\..\\..\\")

##################
### Configure
##################
# The seafile package project directory
# Directory where the signing certificate is located
CERTFILE = "C:/certs/seafile.pfx"

# Qt library directory
QT_DIR = "C:/Qt/6.5.2/msvc2019_64"

# Wix install directory
WIX_BIN = "C:/wix/bin"

# Openssl lib directory
OPENSSL_DIR = "C:/packagelib"

#####################
# Work path : seafile library and program tmp directory
# and wix build path
#####################
# Package directory
SLNOUTPUTDIR = os.path.join(BUILDDIR, "pack")

# Wix package directory
WIX_PACKAGE_DIR = os.path.join(BUILDDIR, "wix_pack")

####################
### Global variables
###################
RETRY_COUNT = 3
error_exit = False
version = ''

####################
### Common helper functions
###################

def highlight(content, is_error=False):
    '''Add ANSI color to content to get it highlighted on terminal'''
    dummy = is_error
    return content
    # if is_error:
    #     return '\x1b[1;31m%s\x1b[m' % content
    # else:
    #     return '\x1b[1;32m%s\x1b[m' % content

def info(msg):
    print(highlight('[INFO] ') + msg)

def error(msg=None, usage=None):
    if msg:
        print(highlight('[ERROR] ') + msg)
    if usage:
        print(usage)
    sys.exit(1)

def find_in_path(prog):
    '''Test whether prog exists in system path'''
    dirs = os.environ['PATH'].split(';')
    for d in dirs:
        if d == '':
            continue
        path = os.path.join(d, prog)
        if os.path.exists(path):
            return path

    return None

def run(cmdline, cwd=None, env=None, suppress_stdout=False, suppress_stderr=False):
    '''Specify a command line string'''
    info('running %s, cwd=%s' % (cmdline, cwd if cwd else os.getcwd()))
    with open(os.devnull, 'w') as devnull:
        if suppress_stdout:
            stdout = devnull
        else:
            stdout = sys.stdout

        if suppress_stderr:
            stderr = devnull
        else:
            stderr = sys.stderr

        proc = subprocess.Popen(cmdline,
                                cwd=cwd,
                                stdout=stdout,
                                stderr=stderr,
                                env=env,
                                shell=True)
        ret = proc.wait()
        if ret != 0:
            global error_exit
            error_exit = True
        return ret

def must_copy(src, dst):
    '''Copy src to dst, exit on failure'''
    try:
        shutil.copy(src, dst)
    except Exception as e:
        error('failed to copy %s to %s: %s' % (src, dst, e))

def must_copytree(src, dst):
    '''Copy dir src to dst, exit on failure'''
    try:
        shutil.copytree(src, dst)
    except Exception as e:
        error('failed to copy dir %s to %s: %s' % (src, dst, e))

def must_rmtree(path):
    '''Recurse rm dir, exit on failure'''
    try:
        shutil.rmtree(path)
    except Exception as e:
        error('failed rm dir %s : %s' % (path, e))

def must_rename(src, dst):
    '''Rename src to dst, exit on failure'''
    try:
        os.rename(src,dst)

    except Exception as e:
        error('failed to rename %s to %s: %s' % (src, dst, e))

def must_mkdir(path):
    '''Creating directories recursively, exit on failure'''
    if os.path.exists(path):
        return
    try:
        os.makedirs(path)
    except OSError as e:
        error('failed to create directory %s:%s' % (path, e))

def dump_env():
    print('Dumping environment variables:')
    for k, v in os.environ.items():
        print('%s: %s' % (k, v)
)

def do_sign(certfile, fn, desc=None):
    info('signing file {} using cert "{}"'.format(fn, certfile))

    if desc:
        desc_flags = '-d "{}"'.format(desc)
    else:
        desc_flags = ''

    signcmd = 'signtool.exe sign -fd sha256 -t http://timestamp.comodoca.com -f {} {} {}'.format(certfile, desc_flags, fn)
    i = 0
    while i < RETRY_COUNT:
        time.sleep(30)
        ret = run(signcmd, cwd=os.path.dirname(fn))
        if ret == 0:
            break
        i = i + 1
        if i == RETRY_COUNT:
            error('Failed to sign file "{}"'.format(fn))

def initworkspace():
    # Clear build file cache
    if os.path.exists(SLNOUTPUTDIR) :
        must_rmtree(SLNOUTPUTDIR)

    # Create a package directory
    must_mkdir(SLNOUTPUTDIR)

def check_project_version(version):
    '''A valid version must be like 1.2.2, 1.3'''
    if not re.match(r'^[0-9]+(\.[0-9]+)+$', version):
        error('%s is not a valid version' % version, usage="vs-build.py 2.0.0")

def check_cmd_para():
    args = sys.argv
    if len(args) != 2:
        error('The number of parameters is incorrect', usage="vs-build.py 2.0.0")
    global version
    version = args[1]
    check_project_version(version)


class Project(object):
    '''Base class for a project'''
    # Probject name, i.e. libseaprc/seafile/seafile-gui
    name = ''

    # A list of shell commands to configure/build the project
    build_commands = []

    def __init__(self):
        self.prefix = BUILDDIR
        self.projdir = os.path.join(self.prefix, self.name)
        self.outdir = os.path.join(self.projdir, 'x64', 'Release')

    def before_build(self):
        '''Hook method to do project-specific stuff before running build commands'''
        pass

    def build(self):
        '''Build the source'''
        self.before_build()
        info('Building %s' % self.name)
        # dump_env()
        for cmd in self.build_commands:
            if run(cmd, cwd=self.projdir) != 0:
                error('error when running command:\n\t%s\n' % cmd)
        self.after_build()

    def after_build(self):
        pass


class Libsearpc(Project):
    name = 'libsearpc'

    def __init__(self):
        Project.__init__(self)
        self.build_commands = [
            'devenv  "%s/libsearpc.sln" /Rebuild "Release|x64"' %(self.projdir),
        ]

    def after_build(self):
        libsearpc_path = os.path.join(self.outdir, 'libsearpc.dll')
        must_copy(libsearpc_path, SLNOUTPUTDIR)


class Seafile(Project):
    name = 'seafile'
    def __init__(self):
        Project.__init__(self)
        self.build_commands = [
            'devenv %s/seafile.sln /Rebuild "Release|x64"' %(self.projdir),
            'devenv %s/msi/custom/seafile_custom.sln /Rebuild "Release|x64"' %(self.projdir),
        ]

    def before_build(self):
        pass

    def after_build(self):

        # Copy seafile dll file to SLNOUTPUTDIR
        dlls = glob.glob(os.path.join(self.outdir, '*.dll'))
        for dll in dlls:
            must_copy(dll, SLNOUTPUTDIR)

        # Copy seafile.exe file to SLNOUTPUTDIR
        must_copy(os.path.join(self.outdir, 'seaf-daemon.exe'), SLNOUTPUTDIR)

        # Generate breakpad symbol
        dump_syms_path = os.path.join(BUILDDIR, 'breakpad', 'src', 'tools', 'windows', 'Release', 'dump_syms.exe')
        pdb_path = os.path.join(self.outdir, 'seaf-daemon.pdb')
        sym_path = os.path.join(self.outdir, 'seaf-daemon.sym')

        cmd = '%s %s > %s' %(dump_syms_path, pdb_path, sym_path)
        if run(cmd, BUILDDIR) != 0:
            error('error when running command:\n\t%s\n' % cmd)


class SeafileGUI(Project):
    name = 'seafile-client'
    target_name = 'seafile-applet.exe'
    def __init__(self):
        Project.__init__(self)
        self.build_commands = [
            'devenv %s/seafile-client.sln /Rebuild "Release|x64"' %(self.projdir) ,
        ]

    def before_build(self):
        pass

    def after_build(self):
        # Copy WinSparkle.dll to SLNOUTPUTDIR
        must_copy(os.path.join(self.projdir, 'third_party', 'WinSparkle-0.5.3', 'x64', 'Release', 'WinSparkle.dll'), SLNOUTPUTDIR)

        # Copy dll to SLNOUTPUTDIR
        dlls = glob.glob(os.path.join(self.outdir, '*.dll'))
        for dll in dlls:
            if not os.path.exists(dll) :
                must_copy(dll, SLNOUTPUTDIR)

        # Copy openssl lib to package dir
        # openssl_lib_path_list = glob.glob(os.path.join(OPENSSL_DIR, '*.dll'))
        # for lib in openssl_lib_path_list :
        #   must_copy(lib, SLNOUTPUTDIR)

        # Copy seafile-applet.exe to SLNOUTPUTDIR
        must_copy(os.path.join(self.outdir, self.target_name), SLNOUTPUTDIR)

        # Use windeloyqt.exe to copy qt resource file and lib
        windeployqt_path = os.path.join(QT_DIR, 'bin', 'windeployqt.exe')
        seafile_exe_path = os.path.join(SLNOUTPUTDIR, self.target_name)
        cmd = "%s --no-compiler-runtime %s" % (windeployqt_path, seafile_exe_path)
        if run(cmd, cwd = SLNOUTPUTDIR) != 0:
            error('error when running command:\n\t%s\n' % cmd)

        # Sign seafile exe
        need_sign_exe = [
            os.path.join(SLNOUTPUTDIR, self.target_name),
            os.path.join(SLNOUTPUTDIR, 'seaf-daemon.exe')
        ]

        for fn in need_sign_exe:
            do_sign(CERTFILE, fn)

        # Generate breakpad symbol
        dump_syms_path = os.path.join(BUILDDIR, 'breakpad', 'src', 'tools', 'windows', 'Release', 'dump_syms.exe')
        pdb_path = os.path.join(self.outdir, 'seafile-applet.pdb')
        sym_path = os.path.join(self.outdir, 'seafile-applet.sym')

        cmd = '%s %s > %s' %(dump_syms_path, pdb_path, sym_path)
        if run(cmd, BUILDDIR) != 0:
            error('error when running command:\n\t%s\n' % cmd)


class SeafileShellExt(Project):
    name = 'seafile-shell-ext'
    def __init__(self):
        Project.__init__(self)
        self.build_commands = [
            'devenv %s/extensions/seafile_ext.sln /Rebuild "Release|x64"' %(self.projdir),
            'devenv %s/shellext-fix/shellext-fix.sln /Rebuild "Release|x64"' %(self.projdir),
        ]
    def before_build(self):
        pass

    def after_build(self):
        # Copy shellext-fix.exe to SLNOUTPUTDIR
        shellext_fix_target = os.path.join(self.projdir, 'shellext-fix', 'x64\\Release', 'shellext-fix.exe')
        must_copy(shellext_fix_target, SLNOUTPUTDIR)

        # Sign seafileext-fix.exe
        do_sign(CERTFILE, os.path.join(SLNOUTPUTDIR, 'shellext-fix.exe'))


def wix_build(language):
    """ Use wix command to build windows msi install package"""

    CULTURE = 'zh-cn'
    LANG_FILE = 'zh_CN.wxl'
    TARGET = 'seafile.msi'

    if language == 'en':
        CULTURE = 'en-us'
        LANG_FILE = 'en_US.wxl'
        TARGET = 'seafile-en.msi'

    CC = '%s/candle.exe' %(WIX_BIN)
    LD = '%s/light.exe' %(WIX_BIN)

    CFLAGS = '-arch "x64" -nologo -ext WixUIExtension -ext WixUtilExtension'
    LDFLAGS  = '-nologo -spdb -ext  WixUIExtension -ext WixUtilExtension' + \
                 ' -loc %s -cultures:%s -sice:ICE80' % (LANG_FILE, CULTURE)


    generator_fragment_cmd = "%s/Paraffin.exe -dir bin -alias bin -gn group_bin \
        fragment.wxs" %(WIX_BIN)
    if run(generator_fragment_cmd, cwd=WIX_PACKAGE_DIR) != 0:
        error('error wherunning command:\n\t%s\n' % generator_fragment_cmd)

    edit_fragment_wxs()

    build_command = [
        '%s %s WixUI_InstallDir_NoLicense.wxs -o WixUI_InstallDir_NoLicense.wixobj'  % (CC, CFLAGS),
        '%s %s MyInstallDirDlg.wxs -o MyInstallDirDlg.wixobj' % (CC, CFLAGS),
        '%s %s fragment.wxs -o fragment.wixobj' % (CC, CFLAGS),
        '%s %s shell.wxs -o shell.wixobj' % (CC, CFLAGS),
        '%s %s seafile.wxs -o seafile.wixobj' % (CC, CFLAGS),
        '%s %s WixUI_InstallDir_NoLicense.wixobj MyInstallDirDlg.wixobj fragment.wixobj shell.wixobj seafile.wixobj -o %s' %(LD, LDFLAGS, TARGET),
    ]
    for cmd in build_command:
        if run(cmd, cwd=WIX_PACKAGE_DIR) != 0:
            error('error when running command:\n\t%s\n' % cmd)

    # Digitally sign the msi package
    msi_path = os.path.join(WIX_PACKAGE_DIR, TARGET)
    signinstaller(msi_path, language)


def prepare_msi():
    if os.path.exists(WIX_PACKAGE_DIR) :
       must_rmtree(WIX_PACKAGE_DIR)


    msi_dir = os.path.join(Seafile().projdir, 'msi')

    # These files are in seafile-shell-ext because they're shared between seafile/seafile
    ext_wxi = os.path.join(SeafileShellExt().projdir, 'msi', 'ext.wxi')
    must_copy(ext_wxi, msi_dir)
    shell_wxs = os.path.join(SeafileShellExt().projdir, 'msi', 'shell.wxs')
    must_copy(shell_wxs, msi_dir)

    # Copy msi to wix package directory
    if not os.path.exists(WIX_PACKAGE_DIR):
        must_copytree(msi_dir, WIX_PACKAGE_DIR)


    wix_pack_bin = os.path.join(WIX_PACKAGE_DIR, 'bin')
    if os.path.exists(wix_pack_bin) :
        os.rmdir(wix_pack_bin)

    # Copy vc runtimer merge module
    must_copy(os.path.join(OPENSSL_DIR, 'Microsoft_VC142_CRT_x64.msm'), SLNOUTPUTDIR)

    must_copytree(SLNOUTPUTDIR, wix_pack_bin)

    # Copy seafile_ext64.dll  to WIX_PACKAGE_DIR/custom
    seafile_extension_target_path = os.path.join(SeafileShellExt().projdir, 'extensions', 'x64\\Release', 'seafile_ext64.dll')
    seafile_extension_dst_path = os.path.join(WIX_PACKAGE_DIR, 'custom')
    must_copy(seafile_extension_target_path, seafile_extension_dst_path)

    # Copy seafile_custom64.dll to WIX_PACKAGE_DIR/custom
    seafile_custom_target_path = os.path.join(Seafile().projdir, 'msi\\custom\\x64\\Release', 'seafile_custom64.dll')
    must_copy(seafile_custom_target_path, seafile_extension_dst_path)


def edit_fragment_wxs():
    '''In the main wxs file(seafile.wxs) we need to reference to the id of
    seafile-applet.exe, which is listed in fragment.wxs. Since fragments.wxs is
    auto generated, the id is sequentially generated, so we need to change the
    id of seafile-applet.exe manually.

    '''
    file_path = os.path.join(WIX_PACKAGE_DIR, 'fragment.wxs')
    new_lines = []
    with open(file_path, 'r', encoding='utf-8') as fp:
        for line in fp:
            if 'seafile-applet.exe' in line:
                # change the id of 'seafile-applet.exe' to 'seafileapplet.exe'
                new_line = re.sub(r'file_[\w]+', 'seafileapplet.exe', line)
                new_lines.append(new_line)
            else:
                new_lines.append(line)

    content = '\r\n'.join(new_lines)
    with open(file_path, 'w', encoding='utf-8') as fp:
        fp.write(content)

def signinstaller(msi_file_path, language):
    global version
    msi_name = ''
    if language != 'cn':
        msi_name = 'seafile-{}-{}.msi' .format(version, language)
    else:
        msi_name = 'seafile-{}.msi' .format(version)
    do_sign(CERTFILE, msi_file_path, msi_name)
    must_rename(msi_file_path, os.path.join(WIX_PACKAGE_DIR, msi_name))


def build_and_sign_msi():
    prepare_msi()

    # Build seafile msi english and chinese version
    wix_build('en')
    wix_build('cn')


def main():
    check_cmd_para()

    # Construct seafile build folder
    initworkspace()

    libsearpc = Libsearpc()
    seafile = Seafile()
    seafile_gui = SeafileGUI()
    seafile_shell_ext = SeafileShellExt()

    # Build Seadrive project
    libsearpc.build()
    seafile.build()
    seafile_gui.build()
    seafile_shell_ext.build()

    # Build seafile msi installer use wix
    build_and_sign_msi()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
#coding: UTF-8
"""Generate the breakpad symbol file and place it in the directory structure
required by breakpad `minidump_stackwalk` tool.

The directory is ./symbols/seaf-daemon.exe/${symbol_id}/seaf-daemon.exe.sym,
where symbol_id is the first line of the "dump_syms" output.
"""

from __future__ import print_function
import os
from os.path import abspath, basename, exists, dirname, join
import re
import sys
import subprocess
import optparse


def call(*a, **kw):
    kw.setdefault('shell', True)
    subprocess.check_call(*a, **kw)


def get_command_output(cmd, **kw):
    shell = not isinstance(cmd, list)
    return subprocess.check_output(cmd, shell=shell, **kw)


def generate_seafile_breakpad_symbol(project_dir, program_name, symbols_output):
    """
    :param project_dir : [string]
    :param program_name : [string]
    :param symbols_output: [string]
    :return: None

    generate_seafile_breakpad_symbol
    """
    os.chdir(project_dir)
    seaf_daemon = join('daemon', '.libs', program_name)
    if not exists(seaf_daemon):
        seaf_daemon = join('daemon', program_name)
        if not exists(seaf_daemon):
            raise RuntimeError('seaf-daemon executable not found!')
    symbols = get_command_output('dump_syms {}'.format(seaf_daemon))

    if symbols_output:
        symbol_file = symbols_output
    else:
        symbol_id = symbols.splitlines()[0].split()[3]
        symbol_dir = join('symbols', program_name, symbol_id)
        if not exists(symbol_dir):
            os.makedirs(symbol_dir)
        symbol_file = join(symbol_dir, '{}.sym'.format(program_name))
    print('symbols written to {}'.format(symbol_file))
    with open(symbol_file, 'w') as fp:
        fp.write(symbols)


def generate_seafile_gui_breakpad_symbol(project_dir, program_name, symbol_output):
    """
    :param project_dir: [string]
    :param program_name: [string]
    :param symbol_output: [string]
    :return: None

     generate seafile gui breakpad symbol
     """

    os.chdir(project_dir)
    seafile_gui_path = os.path.join(project_dir, program_name)
    if not exists(seafile_gui_path):
        raise RuntimeError('seafile gui executable not found !')

    symbols = get_command_output('dump_syms {}'.format(seafile_gui_path))

    with open(symbol_output, 'w') as fp:
        fp.write(symbols)


def main():
    parser = optparse.OptionParser()
    parser.add_option('--projectSrc', help='the project source file directory')
    parser.add_option('--name', help='the program name need to generated breakpad symbol')
    parser.add_option('--output', help='the path of the symbol file.')
    args, _ = parser.parse_args()

    src_dir = args.projectSrc
    program_name = args.name
    symbols_output = args.output
    if program_name == 'seaf-daemon.exe' or program_name == 'seaf-daemon':
        # generate seafile breakpad symbols
        generate_seafile_breakpad_symbol(src_dir, program_name, symbols_output)
    else:
        # generate seafile-gui breakpad symbols
        generate_seafile_gui_breakpad_symbol(src_dir, program_name, symbols_output)


if __name__ == '__main__':
    main()



#coding: utf-8

import os
import time
from . import test_util

def test_add_file():
    test_util.mkfile(1, 'a.md', 'add a file')
    test_util.verify_result()

def test_add_file_t():
    test_util.mkfile(2, 'l/m/n/test.md', 'add l/m/n/test.md')
    test_util.verify_result()

def test_add_dir():
    test_util.mkdir(1, 'ad')
    test_util.verify_result()

def test_add_dir_t():
    test_util.mkdir(2, 'tt/ee/st')
    test_util.verify_result()

def test_modify_file():
    test_util.modfile(1, 'a.md', 'modify a.md')
    test_util.verify_result()

def test_rm_file():
    test_util.rmfile(1, 'a.md')
    test_util.verify_result()

def test_rm_dir():
    test_util.rmdir(1, 'ad')
    test_util.verify_result()

def test_rename_file():
    test_util.mkfile(2, 'b.md', 'add b.md')
    time.sleep(1)
    test_util.move(2, 'b.md', 'b_bak.md')
    test_util.verify_result()

def test_rename_dir():
    test_util.mkdir(2, 'ab')
    time.sleep(1)
    test_util.move(2, 'ab', 'ab_bak')
    test_util.verify_result()

def test_each():
    test_util.mkdir(1, 'abc1')
    test_util.mkfile(1, 'abc1/c.md', 'add abc1/c.md')
    time.sleep(1)

    test_util.mkdir(2, 'bcd1')
    test_util.mkfile(2, 'bcd1/d.md', 'add bcd1/d.md')
    test_util.verify_result()

def test_unsync_resync():
    test_util.desync_cli1()
    test_util.rmdir(1, 'abc1')
    test_util.modfile(1, 'bcd1/d.md', 'modify bcd1/d.md to test unsync resync')
    test_util.sync_cli1()

    test_util.verify_result()

    if not os.path.exists(test_util.getpath(1, 'abc1')):
        assert False, 'dir abc1 should be recreated when resync'

    if len(os.listdir(test_util.getpath(1, 'bcd1'))) != 2:
        assert False, 'should generate conflict file for bcd1/d.md when resync'

def test_modify_timestamp():
    test_util.touch(1, 'bcd1/d.md')
    test_util.verify_result()



#coding: utf-8

import os
import shutil
import time
import subprocess
import glob
from setting import Setting
import seaf_op

def call_process(params):
    with open(os.devnull, 'w') as fd:
        ret = subprocess.check_output(params, stderr=fd)
    return ret

class TestUtil():
    def __init__(self):
        self.setting = Setting()
        self.cli1_dir = os.path.join(os.getcwd(), 'cli1')
        self.cli2_dir = os.path.join(os.getcwd(), 'cli2')
        self.worktree1 = os.path.join(os.getcwd(), 'worktree1')
        self.worktree2 = os.path.join(os.getcwd(), 'worktree2')
        self.enc_repo = False
        try:
            self.enc_repo = bool(os.environ['ENCRYPTED_REPO'])
        except Exception:
            pass
        self.repo_id = None
        self.test_root = ''

    def set_test_root(self, root):
        self.test_root = root

    @staticmethod
    def clean_sync_data(conf):
        try:
            os.remove(os.path.join(conf, 'ccnet.sock'))
            os.remove(os.path.join(conf, 'seafile.ini'))
            shutil.rmtree(os.path.join(conf, 'logs'))
            shutil.rmtree(os.path.join(conf, 'misc'))
            shutil.rmtree(os.path.join(conf, 'seafile'))
            shutil.rmtree(os.path.join(conf, 'seafile-data'))
        except Exception:
            pass

    def init_conf(self):
        self.setting.parse_config()

        if not os.path.exists(self.cli1_dir) or not os.path.exists(self.cli2_dir):
            raise Exception('ccnet conf dir is missing')

        if os.name != 'nt':
            TestUtil.clean_sync_data(self.cli1_dir)
            TestUtil.clean_sync_data(self.cli2_dir)

        if os.path.exists(self.worktree1):
            shutil.rmtree(self.worktree1)
        if os.path.exists(self.worktree2):
            shutil.rmtree(self.worktree2)

        os.mkdir(self.worktree1)
        os.mkdir(self.worktree2)

        seaf_op.seaf_init(self.cli1_dir)
        seaf_op.seaf_init(self.cli2_dir)

    def start_daemon(self):
        seaf_op.seaf_start_all(self.cli1_dir)
        seaf_op.seaf_start_all(self.cli2_dir)

    def create_repo(self):
        self.repo_id = seaf_op.seaf_create(self.cli1_dir, self.setting.server_url,
                                           self.setting.user, self.setting.password,
                                           self.enc_repo)

    def sync_cli1(self):
        seaf_op.seaf_sync(self.cli1_dir, self.setting.server_url, self.repo_id,
                          self.worktree1, self.setting.user, self.setting.password)

    def sync_cli2(self):
        seaf_op.seaf_sync(self.cli2_dir, self.setting.server_url, self.repo_id,
                          self.worktree2, self.setting.user, self.setting.password)

    def sync_repo(self):
        self.sync_cli1()
        self.sync_cli2()

    def desync_cli1(self):
        seaf_op.seaf_desync(self.cli1_dir, self.worktree1)

    def desync_cli2(self):
        seaf_op.seaf_desync(self.cli2_dir, self.worktree2)

    def desync_repo(self):
        self.desync_cli1()
        self.desync_cli2()
        # delete test repo
        seaf_op.seaf_delete(self.cli1_dir, self.setting.server_url,
                            self.setting.user, self.setting.password,
                            self.repo_id)

    def stop_daemon(self):
        seaf_op.seaf_stop(self.cli1_dir)
        seaf_op.seaf_stop(self.cli2_dir)

    def clean(self):
        try:
            if os.name != 'nt':
                TestUtil.clean_sync_data(self.cli1_dir)
                TestUtil.clean_sync_data(self.cli2_dir)
            shutil.rmtree(self.worktree1)
            shutil.rmtree(self.worktree2)
        except Exception:
            pass

    def wait_sync(self):
        while True:
            time.sleep(5)
            repo1 = seaf_op.seaf_get_repo(self.cli1_dir, self.repo_id)
            if repo1 is None:
                continue
            repo2 = seaf_op.seaf_get_repo(self.cli2_dir, self.repo_id)
            if repo2 is None:
                continue
            if repo1.head_cmmt_id == repo2.head_cmmt_id:
                break

    @staticmethod
    def verify_by_rsync(dir1, dir2):
        ret = call_process(['rsync', '-acrin', dir1, dir2])
        if ret:
            for d in ret.split('\n'):
                # omit empty str
                if not d:
                    continue
                # omit directory has almost same result except st_mod
                items = d.split(' ')
                dattr = items[0]
                name = items[1]

                # Output format difference between rsync versions:
                # rsync 3.1.1 : '.d..t.......'
                # rsync 3.0.9 : '.d..t......'

                # On Windows, file timestamp may have 1 second difference
                # between two clients after sync. That's caused by the
                # precision lose when converting Windows timestamp to
                # Unix timestamp. So we don't check timestamp difference
                # for files either.
                if not all([c in ('f', 'd', 't', '.') for c in dattr]):
                    assert False, 'Sync with two client have different result: %s %s' % (dattr, name)

    def verify_result(self, callable=None):
        self.wait_sync()
        if callable:
            callable(self.worktree1, self.worktree2)
        else:
            dir1 = './worktree1/'
            dir2 = './worktree2/'
            TestUtil.verify_by_rsync(dir1, dir2)
            TestUtil.verify_by_rsync(dir2, dir1)

    # worktree: 1(worktree1), 2(worktree2)
    def mkdir(self, worktree, path):
        if worktree == 1:
            os.makedirs(os.path.join(self.worktree1, self.test_root, path))
        elif worktree == 2:
            os.makedirs(os.path.join(self.worktree2, self.test_root, path))

    def rmdir(self, worktree, path):
        if worktree == 1:
            shutil.rmtree(os.path.join(self.worktree1, self.test_root, path))
        elif worktree == 2:
            shutil.rmtree(os.path.join(self.worktree2, self.test_root, path))

    def mkfile(self, worktree, fpath, con=''):
        if worktree == 1:
            pdir = self.worktree1
        elif worktree == 2:
            pdir = self.worktree2
        else:
            return
        abs_path = os.path.join(pdir, self.test_root, fpath)
        dirname = os.path.dirname(abs_path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        with open(abs_path, 'w') as fd:
            fd.write(con)

    def rmfile(self, worktree, fpath):
        if worktree == 1:
            os.remove(os.path.join(self.worktree1, self.test_root, fpath))
        elif worktree == 2:
            os.remove(os.path.join(self.worktree2, self.test_root, fpath))

    def modfile(self, worktree, fpath, con=''):
        if worktree == 1:
            pdir = self.worktree1
        elif worktree == 2:
            pdir = self.worktree2
        else:
            return
        abs_path = os.path.join(pdir, self.test_root, fpath)
        with open(abs_path, 'a') as fd:
            fd.write(con)

    def move(self, worktree, org_path, dest_path):
        if worktree == 1:
            shutil.move(os.path.join(self.worktree1, self.test_root, org_path),
                        os.path.join(self.worktree1, self.test_root, dest_path))
        elif worktree == 2:
            shutil.move(os.path.join(self.worktree2, self.test_root, org_path),
                        os.path.join(self.worktree2, self.test_root, dest_path))

    def batchmove(self, worktree, regex, dest_path):
        if worktree == 1:
            pdir = self.worktree1
        elif worktree == 2:
            pdir = self.worktree2
        else:
            return
        files = glob.glob(os.path.join(pdir, self.test_root, regex))
        dest = os.path.join(pdir, self.test_root, dest_path)
        for f in files:
            shutil.move(f, dest)

    def copy(self, worktree, org_path, dest_path):
        if worktree == 1:
            shutil.copytree(os.path.join(self.worktree1, self.test_root, org_path),
                            os.path.join(self.worktree1, self.test_root, dest_path))
        elif worktree == 2:
            shutil.copytree(os.path.join(self.worktree2, self.test_root, org_path),
                            os.path.join(self.worktree2, self.test_root, dest_path))

    def touch(self, worktree, path, time=None):
        if worktree == 1:
            os.utime(os.path.join(self.worktree1, self.test_root, path), time)
        if worktree == 2:
            os.utime(os.path.join(self.worktree2, self.test_root, path), time)

    def getpath(self, worktree, path):
        if worktree == 1:
            return os.path.join(self.worktree1, self.test_root, path)
        elif worktree == 2:
            return os.path.join(self.worktree2, self.test_root, path)
        raise Exception('Invalid worktree')



from util import TestUtil
import time

test_util = TestUtil()

def setup():
    # init sync related stuff
    test_util.init_conf()
    test_util.start_daemon()
    time.sleep(2)
    test_util.create_repo()
    test_util.sync_repo()
    time.sleep(5)
    print '\n----------------------------------------------------------------------'

def teardown():
    # clean sync related stuf
    print '----------------------------------------------------------------------\n'
    test_util.desync_repo()
    test_util.stop_daemon()
    test_util.clean()



#coding: utf-8

import os
import time
import glob
from threading import Thread
from seaf_op import get_token, urlopen
from . import test_util

'''
Test add-delete-add sequence.
'''
def test_add_delete_add():
    test_util.set_test_root('test_add_delete_add')

    test_util.mkfile(1, 'test/1.txt', 'aaaaaaaa')
    test_util.rmdir(1, 'test')
    test_util.mkfile(1, 'test/1.txt', 'aaaaaaaa')

    test_util.verify_result()

'''
Rename test:
echo 111 > 1.txt; sleep 3; mv 1.txt 2.txt
echo 222 > 3.txt; sleep 3; mv 2.txt 3.txt
echo test > test.txt
mkdir test; echo 444 > 4.txt; sleep 3; mv *.txt test
mkdir test2; mv test test2
mv test2/test .
echo 555 >> test/4.txt; mv test test2
mv test2 test3
'''
def test_rename():
    test_util.set_test_root('test_rename')

    test_util.mkfile(1, '1.txt', '111')
    time.sleep(3)
    test_util.move(1, '1.txt', '2.txt')

    time.sleep(3)

    test_util.mkfile(1, '3.txt', '222')
    time.sleep(3)
    test_util.move(1, '2.txt', '3.txt')

    time.sleep(3)

    test_util.mkfile(1, 'test.txt', 'test')

    time.sleep(3)

    test_util.mkdir(1, 'test')
    test_util.mkfile(1, '4.txt', '444')
    time.sleep(3)
    test_util.batchmove(1, '*.txt', 'test')

    time.sleep(3)

    test_util.mkdir(1, 'test2')
    test_util.move(1, 'test', 'test2')

    time.sleep(3)

    test_util.move(1, 'test2/test', '')

    time.sleep(3)

    test_util.modfile(1, 'test/4.txt', '555')
    test_util.move(1, 'test', 'test2')

    time.sleep(3)

    test_util.move(1, 'test2', 'test3')

    test_util.verify_result()

'''
Create and update test:
echo 111 > test/1.txt
echo 222 >> test/1.txt
copy a dir with multiple levels into test dir
create an empty dir
add file into an empty folder
'''
def test_create_update():
    test_util.set_test_root('test_create_update')

    test_util.mkdir(1, 'test')
    test_util.mkfile(1, 'test/1.txt', '111')

    time.sleep(3)

    test_util.modfile(1, 'test/1.txt', '222')

    time.sleep(3)

    test_util.mkdir(1, '1/2/3/4/5')
    test_util.mkfile(1, '1/1.txt', '111')
    test_util.mkfile(1, '1/2/2.txt', '222')
    test_util.copy(1, '1', 'test/1')

    time.sleep(3)

    test_util.mkdir(1, 'empty')

    time.sleep(3)

    test_util.mkfile(1, 'empty/test.md', 'dddddddddddddddddddddd')

    test_util.verify_result()

'''
Delete test:
echo 222 > 2.txt
rm 2.txt
delete a dir with multiple levels
delete all files under a dir, make it empty.
delete empty dir
'''
def test_delete():
    test_util.set_test_root('test_delete')

    test_util.mkfile(1, '2.txt', '2222')
    time.sleep(3)
    test_util.rmfile(1, '2.txt')

    test_util.mkdir(1, '1/2/3/4/5')
    test_util.mkfile(1, '1/1.txt', '111')
    test_util.mkfile(1, '1/2/2.txt', '222')
    test_util.copy(1, '1', 'test/1')
    time.sleep(3)
    test_util.rmdir(1, '1')

    time.sleep(3)

    test_util.rmdir(1, 'test/1')

    time.sleep(3)

    test_util.rmdir(1, 'test')

    test_util.verify_result()

'''
Case rename test:
rename a dir from 'test' to 'TEST'
disalbe auto sync; rename the dir from 'TEST' to 'test'; enable auto sync.
'''
def test_case_rename():
    test_util.set_test_root('test_case_rename')

    test_util.mkdir(1, 'test')
    test_util.mkfile(1, 'a.txt', 'aaaa')

    time.sleep(3)

    test_util.move(1, 'test', 'TEST')
    test_util.verify_result()

    test_util.desync_cli1()
    test_util.move(1, 'TEST', 'test')
    test_util.sync_cli1()
    test_util.verify_result()

'''
A test set for downloads. The updates are done on cli1 and downloaded on cli2.
Note that these tests are different from upload tests. In upload tests, we deliberately
combine multiple operations into one test; in download tests, we must ensure each
operation be carried out individually on cli2.
'''

# Create a new file
def test_download_1():
    test_util.set_test_root('test_download')

    test_util.mkfile(1, '1.txt', '11111')

    test_util.verify_result()

# Update a file
def test_download_2():
    test_util.set_test_root('test_download')

    test_util.modfile(1, '1.txt', '22222')

    test_util.verify_result()

# Create empty dir
def test_download_3():
    test_util.set_test_root('test_download')

    test_util.mkdir(1, 'dir1')

    test_util.verify_result()

# Rename a file
def test_download_4():
    test_util.set_test_root('test_download')

    test_util.move(1, '1.txt', '2.txt')

    test_util.verify_result()

# Rename empty dir
def test_download_5():
    test_util.set_test_root('test_download')

    test_util.move(1, 'dir1', 'dir2')

    test_util.verify_result()

# Create file in empty dir
def test_download_6():
    test_util.set_test_root('test_download')

    test_util.mkfile(1, 'dir2/1.txt', '1111111')

    test_util.verify_result()

# Rename a non-empty dir
def test_download_7():
    test_util.set_test_root('test_download')

    test_util.move(1, 'dir2', 'dir3')

    test_util.verify_result()

# Remove all files in a non-empty dir
def test_download_8():
    test_util.set_test_root('test_download')

    test_util.rmfile(1, 'dir3/1.txt')

    test_util.verify_result()

# Move a non-empty dir into an empty dir
def test_download_9():
    test_util.set_test_root('test_download')

    test_util.mkfile(1, 'dir4/2.txt', '2222222')
    test_util.verify_result()

    test_util.move(1, 'dir4', 'dir3')

    test_util.verify_result()

# Delete file
def test_download_10():
    test_util.set_test_root('test_download')

    test_util.rmfile(1, '2.txt')

    test_util.verify_result()

# Delete a non-empty dir
def test_download_11():
    test_util.set_test_root('test_download')

    test_util.rmdir(1, 'dir3/dir4')

    test_util.verify_result()

# Delete empty dir
def test_download_12():
    test_util.set_test_root('test_download')

    test_util.rmdir(1, 'dir3')

    test_util.verify_result()

'''
Test cases for download case rename
'''

def test_download_case_rename_1():
    test_util.set_test_root('test_download_case_rename')

    test_util.mkfile(1, 'abc/test/test.txt', 'testtset')
    test_util.verify_result()

    test_util.move(1, 'abc/test', 'abc/TEST')
    test_util.move(1, 'abc', 'ABC')
    test_util.verify_result()

def test_download_case_rename_2():
    test_util.set_test_root('test_download_case_rename')

    test_util.mkfile(1, 'a.txt', 'aaaaaaaaaaaaaaaaaaa')
    test_util.mkfile(1, 'test/b.txt', 'bbbbbbbbbb')
    test_util.verify_result()

    test_util.move(1, 'test', 'TEST')
    test_util.move(1, 'a.txt', 'TEST')
    test_util.verify_result()

'''create cur.md, wait for synced then concurrent modify cur.md
'''
# def mod_file(worktree, fname):
#     test_util.modfile(worktree, fname, test_util.getpath(worktree, fname))

# def test_concurrent_mod():
#     test_util.mkfile(1, 'cur.md', 'curcurcurrrrrrrrrrrrrrrrrrrrr')
#     test_util.verify_result()
#     thread1 = Thread(target=mod_file, args=(1, 'cur.md'))
#     thread2 = Thread(target=mod_file, args=(2, 'cur.md'))
#     thread1.start()
#     thread2.start()
#     test_util.verify_result()
#     files = glob.glob(test_util.getpath(1, 'cur*.md'))
#     assert len(files) ==  2, 'Should generate conflict file'



#!/usr/bin/env python
# coding: UTF-8

'''This script builds the seafile mac client.
'''

import atexit
import logging
import commands
from contextlib import contextmanager
import glob
import multiprocessing
import optparse
import os
from os.path import abspath, basename, dirname, join, exists, expanduser
import re
import shutil
import subprocess
import sys
import tempfile

FINAL_APP = 'Seafile Client.app'
FSPLUGIN_APPEX_NAME = 'Seafile FinderSync.appex'
CERT_ID = '79AB1AF435DD2CBC5FDB3EBBD45B0DA17727B299'

if 'SEAFILE_BUILD_SLAVE' not in os.environ:
    import fabric
    from fabric.api import execute, env, settings, run as fabric_run
    from fabric.operations import put as put_file, get as get_file
    # fabric login configs
    env.user = 'vagrant'
    env.password = 'vagrant'
    env.use_ssh_config = True

####################
### Requires Python 2.6+
####################
if sys.version_info[0] == 3:
    print 'Python 3 not supported yet. Quit now.'
    sys.exit(1)
if sys.version_info[1] < 6:
    print 'Python 2.6 or above is required. Quit now.'
    sys.exit(1)

####################
### Global variables
####################

# command line configuartion
conf = {}

# key names in the conf dictionary.
# pylint: disable=bad-whitespace
CONF_CLEAN              = 'clean'
CONF_MODE               = 'mode'
CONF_VERSION            = 'version'
CONF_SEAFILE_VERSION    = 'seafile_version'
CONF_SEAFILE_CLIENT_VERSION  = 'seafile_client_version'
CONF_LIBSEARPC_VERSION  = 'libsearpc_version'
CONF_SRCDIR             = 'srcdir'
CONF_KEEP               = 'keep'
CONF_BUILDDIR           = 'builddir'
CONF_OUTPUTDIR          = 'outputdir'
CONF_THIRDPARTDIR       = 'thirdpartdir'
CONF_NO_STRIP           = 'nostrip'
CONF_SLAVE_HOST         = 'slave_host'
CONF_BRAND              = 'brand'
CONF_LOCAL              = 'local'
# pylint: enable=bad-whitespace

NUM_CPU = multiprocessing.cpu_count()
PID = os.getpid()

SLAVE_WORKDIR = expanduser('~/tmp/seafile-mac-build-slave')
SLAVE_SRCDIR = join(SLAVE_WORKDIR, 'srcs')
SLAVE_BUILDDIR = join(SLAVE_WORKDIR, 'build')
SLAVE_BUILD_SCRIPT = join(SLAVE_WORKDIR, 'build-mac.py')

####################
### Common helper functions
####################
def highlight(content, is_error=False):
    '''Add ANSI color to content to get it highlighted on terminal'''
    if is_error:
        return '\x1b[1;31m%s\x1b[m' % content
    else:
        return '\x1b[1;32m%s\x1b[m' % content

def info(msg):
    logging.info(highlight('[INFO][{}] ').format(conf.get(CONF_MODE, '')) + msg)

@contextmanager
def cd(path):
    oldpwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(oldpwd)

def exist_in_path(prog):
    '''Test whether prog exists in system path'''
    return bool(find_in_path(prog))

def prepend_env_value(name, value, seperator=':'):
    '''append a new value to a list'''
    try:
        current_value = os.environ[name]
    except KeyError:
        current_value = ''

    new_value = value
    if current_value:
        new_value += seperator + current_value

    os.environ[name] = new_value

def find_in_path(prog):
    '''Find a file in system path'''
    dirs = os.environ['PATH'].split(':')
    for d in dirs:
        if d == '':
            continue
        path = join(d, prog)
        if exists(path):
            return path

    return None

def error(msg=None, usage=None):
    if msg:
        print highlight('[ERROR] ') + msg
    if usage:
        print usage
    sys.exit(1)

def run_argv(argv, cwd=None, env=None, suppress_stdout=False, suppress_stderr=False):
    '''Run a program and wait it to finish, and return its exit code. The
    standard output of this program is supressed.

    '''
    info('running %s, cwd=%s' % (' '.join(['"{}"'.format(a) for a in argv]), cwd or os.getcwd()))
    with open(os.devnull, 'w') as devnull:
        if suppress_stdout:
            stdout = devnull
        else:
            stdout = sys.stdout

        if suppress_stderr:
            stderr = devnull
        else:
            stderr = sys.stderr

        proc = subprocess.Popen(argv,
                                cwd=cwd,
                                stdout=stdout,
                                stderr=stderr,
                                env=env)
        return proc.wait()

def run(cmdline, cwd=None, env=None, suppress_stdout=False, suppress_stderr=False):
    '''Like run_argv but specify a command line string instead of argv'''
    info('running %s, cwd=%s' % (cmdline, cwd or os.getcwd()))
    with open(os.devnull, 'w') as devnull:
        if suppress_stdout:
            stdout = devnull
        else:
            stdout = sys.stdout

        if suppress_stderr:
            stderr = devnull
        else:
            stderr = sys.stderr

        proc = subprocess.Popen(cmdline,
                                cwd=cwd,
                                stdout=stdout,
                                stderr=stderr,
                                env=env,
                                shell=True)
        return proc.wait()

def must_run(cmdline, *a, **kw):
    ret = run(cmdline, *a, **kw)
    if ret != 0:
        error('failed to run %s' % cmdline)

def check_remove(path):
    """
    Remove the file/dir specified by `path`, if exists.
    """
    path = abspath(path)
    assert path.count('/') >= 2
    if exists(path):
        is_dir = os.path.isdir(path)
        info('removing {} {}'.format('dir' if is_dir else 'file', path))
        if is_dir:
            shutil.rmtree(path)
        else:
            os.unlink(path)

def must_mkdir(path):
    '''Create a directory, exit on failure'''
    try:
        if not exists(path):
            os.mkdir(path)
    except OSError, e:
        error('failed to create directory %s:%s' % (path, e))

def must_copy(src, dst):
    '''Copy src to dst, exit on failure'''
    try:
        shutil.copy(src, dst)
    except Exception, e:
        error('failed to copy %s to %s: %s' % (src, dst, e))

def must_copytree(src, dst):
    '''Copy src tree to dst, exit on failure'''
    try:
        shutil.copytree(src, dst)
    except Exception, e:
        error('failed to copy %s to %s: %s' % (src, dst, e))

class Project(object):
    '''Base class for a project'''
    # Probject name, i.e. libseaprc/seafile/
    name = ''

    # A list of shell commands to configure/build the project
    build_commands = []

    def __init__(self):
        # the path to pass to --prefix=/<prefix>
        self.prefix = join(conf[CONF_BUILDDIR], 'usr')
        self.version = self.get_version()
        self.src_tarball = join(conf[CONF_SRCDIR], '%s-%s.tar.gz' % (self.name, self.version))
        # project dir, like <builddir>/seafile-1.2.2/
        self.projdir = join(conf[CONF_BUILDDIR], '%s-%s' % (self.name, self.version))

    def get_version(self):
        # libsearpc can have different versions from seafile.
        raise NotImplementedError

    def get_source_commit_id(self):
        '''By convetion, we record the commit id of the source code in the
        file "<projdir>/latest_commit"

        '''
        latest_commit_file = join(self.projdir, 'latest_commit')
        with open(latest_commit_file, 'r') as fp:
            commit_id = fp.read().strip('\n\r\t ')

        return commit_id

    def append_cflags(self, macros):
        cflags = ' '.join(['-D%s=%s' % (k, macros[k]) for k in macros])
        prepend_env_value('CPPFLAGS',
                          cflags,
                          seperator=' ')

    def uncompress(self):
        '''Uncompress the source from the tarball'''
        info('Uncompressing %s' % self.name)

        if run('tar xf %s' % self.src_tarball) < 0:
            error('failed to uncompress source of %s' % self.name)

    def before_build(self):
        '''Hook method to do project-specific stuff before running build commands'''
        pass

    def build(self):
        '''Build the source'''
        self.before_build()
        info('Building %s' % self.name)
        for cmd in self.build_commands:
            if run(cmd, cwd=self.projdir) != 0:
                error('error when running command:\n\t%s\n' % cmd)

def concurrent_make():
    return 'make -j%s' % NUM_CPU

class Libsearpc(Project):
    name = 'libsearpc'

    def __init__(self):
        Project.__init__(self)
        self.build_commands = [
            './configure --prefix=%s --disable-compile-demo' % self.prefix,
            concurrent_make(),
            'make install'
        ]

    def get_version(self):
        return conf[CONF_LIBSEARPC_VERSION]

class Seafile(Project):
    name = 'seafile'
    def __init__(self):
        Project.__init__(self)
        self.build_commands = [
            './configure --prefix=%s --disable-fuse' % self.prefix,
            concurrent_make(),
            'make install'
        ]

    def get_version(self):
        return conf[CONF_SEAFILE_VERSION]

    def update_cli_version(self):
        '''Substitute the version number in seaf-cli'''
        cli_py = join(self.projdir, 'app', 'seaf-cli')
        with open(cli_py, 'r') as fp:
            lines = fp.readlines()

        ret = []
        for line in lines:
            old = '''SEAF_CLI_VERSION = ""'''
            new = '''SEAF_CLI_VERSION = "%s"''' % conf[CONF_VERSION]
            line = line.replace(old, new)
            ret.append(line)

        with open(cli_py, 'w') as fp:
            fp.writelines(ret)

    def before_build(self):
        self.update_cli_version()
        macros = {}
        # SET SEAFILE_SOURCE_COMMIT_ID, so it can be printed in the log
        macros['SEAFILE_SOURCE_COMMIT_ID'] = '\\"%s\\"' % self.get_source_commit_id()
        self.append_cflags(macros)

class SeafileClient(Project):

    name = 'seafile-client'

    def __init__(self):
        Project.__init__(self)
        cmake_defines = {
            'CMAKE_OSX_ARCHITECTURES': 'x86_64',
            'CMAKE_OSX_DEPLOYMENT_TARGET': '10.9',
            'CMAKE_BUILD_TYPE': 'Release',
            'BUILD_SHIBBOLETH_SUPPORT': 'ON',
            'BUILD_SPARKLE_SUPPORT': 'ON',
        }
        cmake_defines_formatted = ' '.join(['-D{}={}'.format(k, v) for k, v in cmake_defines.iteritems()])
        self.build_commands = [
            'rm -f CMakeCache.txt',
            'cmake -GXcode {}'.format(cmake_defines_formatted),
            'xcodebuild -target seafile-applet -configuration Release -jobs {}'.format(NUM_CPU),
            'rm -rf seafile-applet.app',
            'cp -r Release/seafile-applet.app seafile-applet.app',
            'mkdir -p seafile-applet.app/Contents/Frameworks',
            'macdeployqt seafile-applet.app',
        ]

    def get_version(self):
        return conf[CONF_SEAFILE_CLIENT_VERSION]

    def before_build(self):
        pass

class SeafileDMGLayout(Seafile):

    def __init__(self):
        Seafile.__init__(self)
        self.build_commands = [
        ]

class SeafileFinderSyncPlugin(SeafileClient):

    def __init__(self):
        SeafileClient.__init__(self)
        self.build_commands = [
            'fsplugin/build.sh'
        ]

def check_targz_src(proj, version, srcdir):
    src_tarball = join(srcdir, '%s-%s.tar.gz' % (proj, version))
    if not exists(src_tarball):
        error('%s not exists' % src_tarball)

def validate_args(usage, options):
    required_args = [
        CONF_VERSION,
        CONF_LIBSEARPC_VERSION,
        CONF_SEAFILE_VERSION,
        CONF_SEAFILE_CLIENT_VERSION,
        CONF_SRCDIR,
    ]

    # fist check required args
    for optname in required_args:
        if getattr(options, optname, None) is None:
            error('%s must be specified' % optname, usage=usage)

    def get_option(optname):
        return getattr(options, optname)

    # [ version ]
    def check_project_version(version):
        '''A valid version must be like 1.2.2, 1.3'''
        if not re.match(r'^[0-9]+(\.[0-9]+)+$', version):
            error('%s is not a valid version' % version, usage=usage)

    mode = get_option(CONF_MODE).lower()
    assert mode in ('master', 'slave')

    version = get_option(CONF_VERSION)
    seafile_version = get_option(CONF_SEAFILE_VERSION)
    seafile_client_version = get_option(CONF_SEAFILE_CLIENT_VERSION)
    libsearpc_version = get_option(CONF_LIBSEARPC_VERSION)

    check_project_version(version)
    check_project_version(libsearpc_version)
    check_project_version(seafile_version)
    check_project_version(seafile_client_version)

    # [ srcdir ]
    srcdir = get_option(CONF_SRCDIR)
    check_targz_src('libsearpc', libsearpc_version, srcdir)
    check_targz_src('seafile', seafile_version, srcdir)

    # [ keep ]
    keep = get_option(CONF_KEEP)

    # [ local ]
    build_local = get_option(CONF_LOCAL)

    # [ builddir ]
    builddir = get_option(CONF_BUILDDIR)
    if not exists(builddir):
        error('%s does not exist' % builddir, usage=usage)
    # TODO: remove this
    # if mode == 'master':
    #     builddir = expanduser('~/tmp')
    #     if not exists(builddir):
    #         must_mkdir(builddir)

    builddir = join(builddir, 'seafile-mac-build')

    if not keep:
        check_remove(builddir)
        must_mkdir(builddir)

    # [ outputdir ]
    outputdir = get_option(CONF_OUTPUTDIR)
    if outputdir:
        if not exists(outputdir):
            error('outputdir %s does not exist' % outputdir, usage=usage)
    else:
        outputdir = os.getcwd()

    # [ no strip]
    nostrip = get_option(CONF_NO_STRIP)

    slave_host = get_option(CONF_SLAVE_HOST)

    brand = get_option(CONF_BRAND)

    conf[CONF_MODE] = mode
    conf[CONF_LOCAL] = build_local
    conf[CONF_VERSION] = version
    conf[CONF_LIBSEARPC_VERSION] = libsearpc_version
    conf[CONF_SEAFILE_VERSION] = seafile_version
    conf[CONF_SEAFILE_CLIENT_VERSION] = seafile_client_version

    conf[CONF_BUILDDIR] = builddir
    conf[CONF_SRCDIR] = srcdir
    conf[CONF_OUTPUTDIR] = outputdir
    conf[CONF_KEEP] = keep
    conf[CONF_NO_STRIP] = nostrip
    conf[CONF_SLAVE_HOST] = slave_host
    conf[CONF_BRAND] = brand
    # TODO: remove this
    # conf[CONF_SLAVE_HOST] = 'lion'
    # conf[CONF_SLAVE_HOST] = 'sierra'

    prepare_builddir(builddir)
    # show_build_info()

def show_build_info():
    '''Print all conf information. Confirm before continue.'''
    info('------------------------------------------')
    info('Seafile command line client %s: BUILD INFO' % conf[CONF_VERSION])
    info('------------------------------------------')
    info('libsearpc:        %s' % conf[CONF_LIBSEARPC_VERSION])
    info('seafile:          %s' % conf[CONF_SEAFILE_VERSION])
    info('seafile-client:   %s' % conf[CONF_SEAFILE_CLIENT_VERSION])
    info('builddir:         %s' % conf[CONF_BUILDDIR])
    info('outputdir:        %s' % conf[CONF_OUTPUTDIR])
    info('source dir:       %s' % conf[CONF_SRCDIR])
    info('strip symbols:    %s' % (not conf[CONF_NO_STRIP]))
    info('clean on exit:    %s' % (not conf[CONF_KEEP]))
    info('------------------------------------------')
    info('press any key to continue ')
    info('------------------------------------------')
    dummy = raw_input()

def prepare_builddir(builddir):
    must_mkdir(builddir)

    # if not conf[CONF_KEEP]:
    #     def remove_builddir():
    #         '''Remove the builddir when exit'''
    #         info('remove builddir before exit')
    #         shutil.rmtree(builddir, ignore_errors=True)
    #     atexit.register(remove_builddir)

    os.chdir(builddir)

    must_mkdir(join(builddir, 'usr'))

def parse_args():
    parser = optparse.OptionParser()
    def long_opt(opt):
        return '--' + opt

    parser.add_option(long_opt(CONF_MODE),
                      dest=CONF_MODE,
                      default='master',
                      help='''The mode: master or slave. Master is osx 10.12, while slave is used when running on osx 10.7 lion.
                      Master would ssh on to slave to compile the source code.''')

    parser.add_option(long_opt(CONF_VERSION),
                      dest=CONF_VERSION,
                      nargs=1,
                      help='the version to build. Must be digits delimited by dots, like 1.3.0')

    parser.add_option(long_opt(CONF_SEAFILE_VERSION),
                      dest=CONF_SEAFILE_VERSION,
                      nargs=1,
                      help='the version of seafile as specified in its "configure.ac". Must be digits delimited by dots, like 1.3.0')

    parser.add_option(long_opt(CONF_LIBSEARPC_VERSION),
                      dest=CONF_LIBSEARPC_VERSION,
                      nargs=1,
                      help='the version of libsearpc as specified in its "configure.ac". Must be digits delimited by dots, like 1.3.0')

    parser.add_option(long_opt(CONF_SEAFILE_CLIENT_VERSION),
                      dest=CONF_SEAFILE_CLIENT_VERSION,
                      nargs=1,
                      help='the version of seafile-client. Must be digits delimited by dots, like 1.3.0')

    parser.add_option(long_opt(CONF_BUILDDIR),
                      dest=CONF_BUILDDIR,
                      nargs=1,
                      help='the directory to build the source. Defaults to /tmp',
                      default=tempfile.gettempdir())

    parser.add_option(long_opt(CONF_OUTPUTDIR),
                      dest=CONF_OUTPUTDIR,
                      nargs=1,
                      help='the output directory to put the generated tarball. Defaults to the current directory.',
                      default=os.getcwd())

    parser.add_option(long_opt(CONF_SRCDIR),
                      dest=CONF_SRCDIR,
                      nargs=1,
                      help='''Source tarballs must be placed in this directory.''')

    parser.add_option(long_opt(CONF_KEEP),
                      dest=CONF_KEEP,
                      action='store_true',
                      help='''keep the build directory after the script exits. By default, the script would delete the build directory at exit.''')

    parser.add_option(long_opt(CONF_NO_STRIP),
                      dest=CONF_NO_STRIP,
                      action='store_true',
                      help='''do not strip debug symbols''')

    parser.add_option(long_opt(CONF_SLAVE_HOST),
                      dest=CONF_SLAVE_HOST,
                      default='lion',
                      help='the hostname of the lower version osx slave')

    parser.add_option(long_opt(CONF_BRAND),
                      dest=CONF_BRAND,
                      help='the brand')

    parser.add_option(long_opt(CONF_LOCAL),
                      dest=CONF_LOCAL,
                      action='store_true',
                      help='build locally instead of using an old 10.7 vm')
    usage = parser.format_help()
    options, remain = parser.parse_args()
    if remain:
        error(usage=usage)

    validate_args(usage, options)

def setup_build_env():
    '''Setup environment variables, such as export PATH=$BUILDDDIR/bin:$PATH'''
    # os.environ.update({
    # })
    prefix = join(conf[CONF_BUILDDIR], 'usr')

    prepend_env_value('CFLAGS',
                      '-Wall -O2 -g -DNDEBUG -I/opt/local/include -mmacosx-version-min=10.7',
                      seperator=' ')

    prepend_env_value('CXXFLAGS',
                      '-Wall -O2 -g -DNDEBUG -I/opt/local/include -mmacosx-version-min=10.7',
                      seperator=' ')

    prepend_env_value('CPPFLAGS',
                      '-I%s' % join(prefix, 'include'),
                      seperator=' ')

    prepend_env_value('CPPFLAGS',
                      '-DSEAFILE_CLIENT_VERSION=\\"%s\\"' % conf[CONF_VERSION],
                      seperator=' ')

    if conf[CONF_NO_STRIP]:
        prepend_env_value('CPPFLAGS',
                          '-g -O0',
                          seperator=' ')

    prepend_env_value('LDFLAGS',
                      '-L%s' % join(prefix, 'lib'),
                      seperator=' ')

    prepend_env_value('LDFLAGS',
                      '-L%s' % join(prefix, 'lib64'),
                      seperator=' ')

    prepend_env_value('LDFLAGS',
                      '-L/opt/local/lib -Wl,-headerpad_max_install_names -mmacosx-version-min=10.7',
                      seperator=' ')

    prepend_env_value('PATH', join(prefix, 'bin'))
    prepend_env_value('PKG_CONFIG_PATH', join(prefix, 'lib', 'pkgconfig'))
    prepend_env_value('PKG_CONFIG_PATH', join(prefix, 'lib64', 'pkgconfig'))

path_pattern = re.compile(r'^\s+(\S+)\s+\(compatibility.*')
def get_dependent_libs(executable):
    def is_syslib(lib):
        if lib.startswith('/usr/lib'):
            return True
        if lib.startswith('/System/'):
            return True
        return False

    otool_output = commands.getoutput('otool -L %s' % executable)
    libs = set()
    for line in otool_output.splitlines():
        m = path_pattern.match(line)
        if not m:
            continue
        path = m.group(1)
        if not is_syslib(path):
            libs.add(path)
    return libs

def copy_shared_libs():
    '''copy shared c libs, such as libevent, glib'''
    builddir = conf[CONF_BUILDDIR]
    frameworks_dir = join(SeafileClient().projdir, 'seafile-applet.app/Contents/Frameworks')

    must_mkdir(frameworks_dir)
    seafile_path = join(builddir, 'usr/bin/seaf-daemon')

    libs = set()
    libs.update(get_dependent_libs(seafile_path))

    # Get deps of deps recursively until no more deps can be included
    while True:
        newlibs = set(libs)
        for lib in libs:
            newlibs.update(get_dependent_libs(lib))
        if newlibs == libs:
            break
        libs = newlibs

    for lib in libs:
        dst_file = join(frameworks_dir, basename(lib))
        if exists(dst_file):
            continue
        info('Copying %s' % lib)
        shutil.copy(lib, frameworks_dir)

    change_rpaths()

def change_rpaths():
    """
    Chagne the rpath of the referenced dylibs so the app can be relocated
    anywhere in the user's system.
    See:
      - https://blogs.oracle.com/dipol/entry/dynamic_libraries_rpath_and_mac
      - https://developer.apple.com/library/content/documentation/DeveloperTools/Conceptual/DynamicLibraries/100-Articles/RunpathDependentLibraries.html
    """
    contents_dir = join(SeafileClient().projdir, 'seafile-applet.app/Contents')
    frameworks_dir = join(contents_dir, 'Frameworks')
    resources_dir = join(contents_dir, 'Resources')
    macos_dir = join(contents_dir, 'MacOS')

    seafile_applet = join(macos_dir, 'seafile-applet')
    binaries = [
        seafile_applet,
        join(resources_dir, 'seaf-daemon')
    ]

    RPATH_RE = re.compile(r'^path\s+(\S+)\s+\(offset .*$')
    def get_rpaths(fn):
        rpaths = []
        output = commands.getoutput('otool -l {} | grep -A2 RPATH || true'.format(fn))
        # The output is like
            #           cmd LC_RPATH
            #       cmdsize 24
            #          path /usr/local (offset 12)
            # --
            #           cmd LC_RPATH
            #       cmdsize 48
            #          path @executable_path/../Frameworks (offset 12)
        for line in output.splitlines():
            m = RPATH_RE.match(line.strip())
            if m:
                rpaths.append(m.group(1))
        return rpaths

    def has_rpath(fn, path):
        return path in get_rpaths(fn)

    def add_frameworks_dir_to_rpath(fn, executable=True):
        relpath = 'executable_path' if executable else 'loader_path'
        relative_frameworks_dir = '@{}/../Frameworks'.format(relpath)
        if not has_rpath(fn, relative_frameworks_dir):
            must_run('install_name_tool -add_rpath {} {}' .format(relative_frameworks_dir, fn))

    def remove_local_rpaths(fn):
        local_paths = ['/usr/local/lib', '/opt/local/lib']
        for path in local_paths:
            if has_rpath(fn, path):
                must_run('install_name_tool -delete_rpath {} {}'.format(path, fn))

    def change_deps_rpath(fn):
        deps = get_dependent_libs(fn)
        for dep in deps:
            bn = basename(dep)
            if 'framework' in bn or bn.startswith('Qt') or bn == 'Sparkle':
                continue
            must_run('install_name_tool -change {} @rpath/{} {}'.format(dep, basename(dep), fn))

    for binary in binaries:
        add_frameworks_dir_to_rpath(binary)
        remove_local_rpaths(binary)
        change_deps_rpath(binary)

    libs = os.listdir(frameworks_dir)
    for lib_name in libs:
        lib = join(frameworks_dir, lib_name)
        if os.path.isdir(lib):
            continue
        must_run('install_name_tool -id @rpath/{} {}'.format(basename(lib), lib))
        add_frameworks_dir_to_rpath(lib, executable=False)
        change_deps_rpath(lib)
        remove_local_rpaths(lib)

DROPDMG = '/Applications/DropDMG.app/Contents/Frameworks/DropDMGFramework.framework/Versions/A/dropdmg'

def gen_dmg():
    output_dmg = 'app-{}.dmg'.format(conf[CONF_VERSION])
    parentdir = 'app-{}'.format(conf[CONF_VERSION])
    appdir = join(parentdir, 'seafile-applet.app')
    app_plugins_dir = join(appdir, 'Contents/PlugIns')

    layout = SeafileDMGLayout()
    layout.uncompress()
    layout_folder = join(layout.projdir, 'dmg/seafileLayout')

    args = [
        DROPDMG,
        parentdir,
        '--format', 'bzip2',
        '--layout-folder', layout_folder,
        '--volume-name', conf[CONF_BRAND] or 'Seafile Client',
    ]

    with cd(conf[CONF_BUILDDIR]):
        check_remove(parentdir)
        check_remove(output_dmg)
        must_mkdir(parentdir)
        must_run('tar xf seafile-applet.app.tar.gz -C {}'.format(parentdir))
        # Remove the Qt Bearer plugin because it would run a background thread
        # to scan wifi networks. See
        # https://trello.com/c/j28eOIo1/359-explicitly-exclude-qt-bearer-plugins-for-the-windows-and-mac-packages
        # for details.
        must_run('rm -rf "{}"'.format(join(app_plugins_dir, 'bearer')))
        # fsplugin must be copied before we sign the final .app dir
        copy_fsplugin(app_plugins_dir)
        sign_files(appdir)

        # Rename the .app dir to 'Seafile Client.app', and create the shortcut
        # to '/Applications' so the user can drag into it when opening the DMG.
        brand = conf.get(CONF_BRAND, '')
        if brand:
            final_app = '{}.app'.format(brand)
        else:
            final_app = FINAL_APP
        must_run('mv {}/seafile-applet.app "{}/{}"'.format(parentdir, parentdir, final_app))
        must_run('ln -sf /Applications {}/'.format(parentdir))

        # Open DropDMG manually, or dropdmg command line may fail.
        run(''' osascript -e 'tell application "DropDMG" to quit' || true ''')
        run(''' osascript -e 'open application "DropDMG"' || true ''')
        run(''' osascript -e 'activate application "DropDMG"' || true ''')

        # Sometimes DropDmg would fail if there are two many Finder windows.
        run(''' osascript -e 'tell application "Finder" to close every window' ''')
        if run_argv(args) != 0:
            error('failed to run {}'.format(args))

def sign_in_parallel(files_to_sign):
    import threading
    import Queue
    queue = Queue.Queue()
    POISON_PILL = ''

    class SignThread(threading.Thread):
        def __init__(self, index):
            threading.Thread.__init__(self)
            self.index = index

        def run(self):
            info('sign thread {} started'.format(self.index))
            while True:
                try:
                    fn = queue.get(timeout=1)
                except Queue.Empty:
                    continue
                else:
                    if fn == POISON_PILL:
                        break
                    else:
                        do_sign(fn)
            info('sign thread {} stopped'.format(self.index))

    TOTAL_THREADS = max(NUM_CPU / 2, 1)
    threads = []
    for i in xrange(TOTAL_THREADS):
        t = SignThread(i)
        t.start()
        threads.append(t)

    for fn in files_to_sign:
        queue.put(fn)

    for _ in xrange(TOTAL_THREADS):
        queue.put(POISON_PILL)

    for i in xrange(TOTAL_THREADS):
        threads[i].join()

def sign_files(appdir):
    webengine_app = join(
        appdir,
        'Contents/Frameworks/QtWebEngineCore.framework/Versions/5/Helpers/QtWebEngineProcess.app'
    )

    def _glob(pattern, *a, **kw):
        return glob.glob(join(appdir, pattern), *a, **kw)

    # The webengine app must be signed first, otherwise the sign of
    # QtWebengineCore.framework would fail.
    if exists(webengine_app):
        entitlements = join(Seafile().projdir, 'scripts/build/osx.entitlements')
        do_sign(
            webengine_app,
            extra_args=['--entitlements', entitlements]
        )

    # Strip the get-task-allow entitlements for Sparkle binaries
    for fn in _glob('Contents/Frameworks/Sparkle.framework/Versions/A/Resources/Autoupdate.app/Contents/MacOS/*'):
        do_sign(fn, preserve_entitlemenets=False)

    # Sign the nested contents of Sparkle before we sign
    # Sparkle.Framework in the thread pool.
    for fn in (
            'Contents/Frameworks/Sparkle.framework/Versions/A/Resources/Autoupdate.app',
            'Contents/Frameworks/Sparkle.framework/Versions/A/Sparkle',
    ):
        do_sign(join(appdir, fn))

    patterns = [
        'Contents/Frameworks/*.framework',
        'Contents/PlugIns/*/*.dylib',
        'Contents/Frameworks/*.dylib',
        'Contents/Resources/seaf-daemon',
        'Contents/MacOS/seadrive-gui',
    ]

    files_to_sign = []
    for p in patterns:
        files_to_sign.extend(_glob(p))

    info('{} files to sign'.format(len(files_to_sign)))

    sign_in_parallel(files_to_sign)
    # for fn in files_to_sign:
    #     do_sign(fn)

    do_sign(appdir)
    # do_sign(appdir, extra_args=['--deep'])

_keychain_unlocked = False
def unlock_keychain():
    """
    Unlock the keychain when we're using ssh instead of using the terminal from
    GUI. See http://stackoverflow.com/a/20208104/1467959
    """
    global _keychain_unlocked
    if not _keychain_unlocked:
        _keychain_unlocked = True
        run('security -v unlock-keychain -p vagrant || true')

def do_sign(path, extra_args=None, preserve_entitlemenets=True):
    unlock_keychain()
    args = [
        'codesign',
        '--verbose=4',
        '-o', 'runtime',
        '--timestamp',
        '--verify',
        # '--no-strict',
        '--force',
        '-s', CERT_ID,
    ]
    if preserve_entitlemenets:
        args += ['--preserve-metadata=entitlements']
    extra_args = extra_args or []
    if extra_args:
        args.extend(extra_args)

    args.append(path)

    if run_argv(args) != 0:
        error('failed to sign {}'.format(path))

def send_file_to_slave(src, dst):
    info('putting file {} => {}'.format(src, dst))
    with settings(host_string=conf[CONF_SLAVE_HOST]):
        put_file(local_path=src, remote_path=dst)

def get_file_from_slave(src, dst):
    host = conf[CONF_SLAVE_HOST]
    info('getting file {}:{} => {}'.format(host, src, dst))
    with settings(host_string=host):
        get_file(src, dst)

def send_project_tarball(project):
    info('sending tarball of {}'.format(project.name))
    src = project.src_tarball
    dst = join(SLAVE_SRCDIR, basename(src))
    send_file_to_slave(src, dst)

def slave_run(cmd, **kw):
    info('running cmd: ' + cmd)
    kw['host'] = conf[CONF_SLAVE_HOST]
    cmd = '. ~/.bash_profile; export SEAFILE_BUILD_SLAVE=1; ' + cmd
    execute(fabric_run, cmd, **kw)

def prepare_dirs_on_slave():
    info('creating work directory on the slave')
    slave_run('mkdir -p {}'.format(SLAVE_SRCDIR))
    slave_run('mkdir -p {}'.format(SLAVE_BUILDDIR))

def send_sources_to_slave():
    info('sending sources to the slave')
    prepare_dirs_on_slave()
    send_file_to_slave(abspath(__file__), SLAVE_BUILD_SCRIPT)
    projects = [
        Libsearpc(),
        Seafile(),
        SeafileClient(),
    ]
    for p in projects:
        send_project_tarball(p)

def build_on_slave():
    args = [
        'python',
        SLAVE_BUILD_SCRIPT,
        '--mode=slave',
        '--version={}'.format(conf[CONF_VERSION]),
        '--libsearpc_version={}'.format(conf[CONF_LIBSEARPC_VERSION]),
        '--seafile_version={}'.format(conf[CONF_SEAFILE_VERSION]),
        '--seafile_client_version={}'.format(conf[CONF_SEAFILE_CLIENT_VERSION]),
        '--srcdir={}'.format(SLAVE_SRCDIR),
        '--builddir={}' .format(SLAVE_BUILDDIR),
    ]
    info('packaging on the slave')
    slave_run(' '.join(args))

def get_app_tgz_from_slave():
    fn = 'seafile-applet.app.tar.gz'
    src = join(SLAVE_BUILDDIR, 'seafile-mac-build', fn)
    dst = join(conf[CONF_BUILDDIR], fn)
    get_file_from_slave(src, dst)

def copy_dmg():
    brand = conf[CONF_BRAND] or 'seafile-client'
    branded_dmg = '{}-{}.dmg'.format(brand, conf[CONF_VERSION])
    src_dmg = os.path.join(conf[CONF_BUILDDIR], 'app-{}.dmg'.format(conf[CONF_VERSION]))
    dst_dmg = os.path.join(conf[CONF_OUTPUTDIR], branded_dmg)

    # move msi to outputdir
    must_copy(src_dmg, dst_dmg)

    print '---------------------------------------------'
    print 'The build is successfully. Output is:'
    print '>>\t%s' % dst_dmg
    print '---------------------------------------------'

def notarize_dmg():
    pkg = os.path.join(conf[CONF_BUILDDIR], 'app-{}.dmg'.format(conf[CONF_VERSION]))
    info('Try to notarize {}'.format(pkg))
    notarize_script = join(Seafile().projdir, 'scripts/build/notarize.sh')
    cmdline = '{} {}'.format(notarize_script, pkg)
    ret = run(cmdline)
    if ret != 0:
        error('failed to notarize: %s' % cmdline)
    info('Successfully notarized {}'.format(pkg))

def build_and_sign_fsplugin():
    """
    Build and sign the fsplugin. The final output would be "${buildder}/Seafile FinderSync.appex"
    """
    fsplugin = SeafileFinderSyncPlugin()
    fsplugin.uncompress()
    fsplugin.build()
    with cd(fsplugin.projdir):
        appex_src = 'fsplugin/{}'.format(FSPLUGIN_APPEX_NAME)
        appex_dst = join(conf[CONF_BUILDDIR], basename(appex_src))

        check_remove(appex_dst)
        must_copytree(appex_src, appex_dst)

        entitlements_src = 'fsplugin/seafile-fsplugin.entitlements'
        entitlements_dst = join(conf[CONF_BUILDDIR], basename(entitlements_src))

        check_remove(entitlements_dst)
        must_copy(entitlements_src, entitlements_dst)

    do_sign(
        appex_dst,
        extra_args=['--entitlements', entitlements_dst]
    )

def copy_fsplugin(plugins_dir):
    src = join(conf[CONF_BUILDDIR], FSPLUGIN_APPEX_NAME)
    dst = join(plugins_dir, FSPLUGIN_APPEX_NAME)
    check_remove(dst)
    must_copytree(src, dst)

def copy_sparkle_framework():
    src = '/usr/local/Sparkle.framework'
    dst = join(SeafileClient().projdir, 'seafile-applet.app/Contents/Frameworks', basename(src))
    check_remove(dst)
    # Here we use the `cp` command instead of shutil to do the copy, because `cp
    # -P` would keep symlinks as is.
    must_run('cp -R -P "{}" "{}"'.format(src, dst))

def build_projects():
    libsearpc = Libsearpc()
    seafile = Seafile()
    seafile_client = SeafileClient()

    libsearpc.uncompress()
    libsearpc.build()

    seafile.uncompress()
    seafile.build()

    seafile_client.uncompress()
    seafile_client.build()

    copy_sparkle_framework()

    copy_shared_libs()

def local_workflow():
    build_projects()
    generate_app_tar_gz()

    build_and_sign_fsplugin()
    gen_dmg()
    notarize_dmg()
    copy_dmg()

def master_workflow():
    send_sources_to_slave()
    build_on_slave()
    get_app_tgz_from_slave()

    build_and_sign_fsplugin()
    gen_dmg()
    notarize_dmg()
    copy_dmg()

def slave_workflow():
    build_projects()
    generate_app_tar_gz()

def generate_app_tar_gz():
    # output_app_tgz = join(conf[CONF_BUILDDIR], '..', 'seafile-applet.app.tar.gz')
    output_app_tgz = join(conf[CONF_BUILDDIR], 'seafile-applet.app.tar.gz')
    with cd(SeafileClient().projdir):
        run('tar czf {} seafile-applet.app'.format(output_app_tgz))

def setup_logging(level=logging.INFO):
    kw = {
        'format': '[%(asctime)s][%(module)s]: %(message)s',
        'datefmt': '%m/%d/%Y %H:%M:%S',
        'level': level,
        'stream': sys.stdout
    }

    logging.basicConfig(**kw)
    logging.getLogger('requests.packages.urllib3.connectionpool').setLevel(
        logging.WARNING)

def main():
    setup_logging()
    parse_args()
    info('{} script started'.format(abspath(__file__)))
    info('NUM_CPU = {}'.format(NUM_CPU))
    setup_build_env()

    if conf[CONF_LOCAL]:
        local_workflow()
    elif conf[CONF_MODE] == 'master':
        # info('entering master workflow')
        # master_workflow()
        local_workflow()
    else:
        info('entering slave workflow')
        slave_workflow()

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding: UTF-8

'''This scirpt is used to bundle all needed files into the destdir to
faciliate windows development.

'''

import sys

####################
### Requires Python 2.6+
####################
if sys.version_info[0] == 3:
    print 'Python 3 not supported yet. Quit now.'
    sys.exit(1)
if sys.version_info[1] < 6:
    print 'Python 2.6 or above is required. Quit now.'
    sys.exit(1)

import os
import glob
import shutil
import subprocess
import tempfile
import csv

from distutils.core import setup as dist_setup
import py2exe

def usage():
    print '''\
Usage:
    %s <target directory>
'''

####################
### Common helper functions
####################
def to_mingw_path(path):
    if len(path) < 2 or path[1] != ':' :
        return path.replace('\\', '/')

    drive = path[0]
    return '/%s%s' % (drive.lower(), path[2:].replace('\\', '/'))

def to_win_path(path):
    if len(path) < 2 or path[1] == ':' :
        return path.replace('/', '\\')

    drive = path[1]
    return '%s:%s' % (drive.lower(), path[2:].replace('/', '\\'))

def info(msg):
    print '[INFO] ' + msg

def error(msg=None, usage=None):
    if msg:
        print '[ERROR] ' + msg
    if usage:
        print usage
    sys.exit(1)

def which(prog):
    '''Return the path of the file <prog>, if exists in PATH'''
    dirs = os.environ['PATH'].split(';')
    for d in dirs:
        if d == '':
            continue
        path = os.path.join(d, prog)
        if os.path.exists(path):
            return path

    return None

def run(cmdline, cwd=None, env=None, suppress_stdout=False, suppress_stderr=False):
    '''Run a program and wait it to finish, and return its exit code. The
    standard output of this program is supressed.

    '''
    info('running %s, cwd=%s' % (cmdline, cwd if cwd else os.getcwd()))
    with open(os.devnull, 'w') as devnull:
        if suppress_stdout:
            stdout = devnull
        else:
            stdout = sys.stdout

        if suppress_stderr:
            stderr = devnull
        else:
            stderr = sys.stderr

        proc = subprocess.Popen(cmdline,
                                cwd=cwd,
                                stdout=stdout,
                                stderr=stderr,
                                env=env,
                                shell=True)
        return proc.wait()

def rmtree(path):
    '''Remove a directory, ignore errors'''
    try:
        shutil.rmtree(path)
    except:
        pass

def must_mkdir(path):
    '''Create a directory, exit on failure'''
    try:
        os.mkdir(path)
    except OSError, e:
        error('failed to create directory %s:%s' % (path, e))

def must_copy(src, dst):
    '''Copy src to dst, exit on failure'''
    try:
        info('copying %s --> %s' % (src, dst))
        shutil.copy(src, dst)
    except Exception, e:
        error('failed to copy %s to %s: %s' % (src, dst, e))

def must_copytree(src, dst):
    '''Copy dir src to dst, exit on failure'''
    try:
        info('copying directory %s --> %s' % (src, dst))
        shutil.copytree(src, dst)
    except Exception, e:
        error('failed to copy dir %s to %s: %s' % (src, dst, e))

def must_move(src, dst):
    '''Move src to dst, exit on failure'''
    try:
        info('moving %s --> %s' % (src, dst))
        shutil.move(src, dst)
    except Exception, e:
        error('failed to move %s to %s: %s' % (src, dst, e))

def web_py2exe():
    webdir = os.path.join(seafile_srcdir, 'web')
    dist_dir = os.path.join(webdir, 'dist')
    build_dir = os.path.join(webdir, 'build')

    rmtree(dist_dir)
    rmtree(build_dir)
    
    os.chdir(webdir)

    original_argv = sys.argv
    sys.argv = [sys.argv[0], 'py2exe']
    sys.path.insert(0, webdir)

    targetname = 'seafile-web'
    targetfile = targetname + '.py'
    must_copy('main.py', targetfile)

    packages=["mako.cache", "utils"]
    ex_files=[]
    option = {"py2exe":
              {"includes" :[targetname],
               "packages" : packages,
               "bundle_files" : 3}}

    try:
        dist_setup(name=targetname,
                   options = option,
                   windows=[{"script":targetfile}],
                   data_files=ex_files)
    except Exception as e:
        error('Error when calling py2exe: %s' % e)

    for name in glob.glob('dist/*'):
        must_copy(name, bin_dir)

    must_copytree('i18n', os.path.join(bin_dir, 'i18n'))
    must_copytree('static', os.path.join(bin_dir, 'static'))
    must_copytree('templates', os.path.join(bin_dir, 'templates'))

    rmtree(dist_dir)
    rmtree(build_dir)

    sys.path.pop(0)
    sys.argv = original_argv
    os.chdir(seafile_srcdir)

def parse_depends_csv(path):
    '''parse the output of dependency walker'''
    libs = []
    def should_ingore_lib(lib):
        if not os.path.exists(lib):
            return True

        if lib.lower().startswith('c:\\windows'):
            return True

        return False

    with open(path, 'r') as fp:
        reader = csv.reader(fp)
        for row in reader:
            if len(row) < 2:
                continue
            lib = row[1]
            if not should_ingore_lib(lib):
                libs.append(lib)

    return libs

def copy_shared_libs():
    '''Copy shared libs need by libccnet, such as libevent, libsqlite, etc.
    First we use Dependency walker to analyse libccnet-0.dll, and get an
    output file in csv format. Then we parse the csv file to get the list of
    shared libs.

    '''

    tempdir = tempfile.gettempdir()
    output = os.path.join(tempdir, 'depends.csv')
    applet = os.path.join(seafile_srcdir, 'gui', 'win', 'seafile-applet.exe')
    cmd = 'depends.exe -c -f 1 -oc %s %s' % (output, applet)

    # See the manual of Dependency walker
    if run(cmd) > 0x100:
        error('failed to run dependency walker for libccnet')

    if not os.path.exists(output):
        error('failed to run dependency walker for libccnet')

    shared_libs = parse_depends_csv(output)
    for lib in shared_libs:
        must_copy(lib, bin_dir)

    libsqlite3 = which('libsqlite3-0.dll')
    must_copy(libsqlite3, bin_dir)

def copy_dll_exe():
    filelist = [
        'libsearpc-1.dll',
        'libsearpc-json-glib-0.dll',
        'libccnet-0.dll',
        'libseafile-0.dll',
        'ccnet.exe',
        'seaf-daemon.exe',
    ]

    filelist = [ which(f) for f in filelist ]

    applet = os.path.join(seafile_srcdir, 'gui', 'win', 'seafile-applet.exe')
    filelist.append(applet)

    for name in filelist:
        must_copy(name, bin_dir)

    copy_shared_libs()

def main():
    if not os.path.exists(destdir):
        must_mkdir(destdir)
    if not os.path.exists(bin_dir):
        must_mkdir(bin_dir)
    web_py2exe()
    copy_dll_exe()

if __name__ == '__main__':
    if len(sys.argv) != 2:
        usage()
        exit(1)

    seafile_srcdir = os.getcwd()
    destdir = to_win_path(sys.argv[1])
    bin_dir = os.path.join(destdir, 'bin')
    main()



from pysearpc import searpc_func, SearpcError, NamedPipeClient


class SeafileRpcClient(NamedPipeClient):
    """RPC used in client"""

    def __init__(self, socket_path, *args, **kwargs):
         NamedPipeClient.__init__(
             self,
             socket_path,
             "seafile-rpcserver",
             *args,
             **kwargs
         )

    @searpc_func("string", ["int"])
    def seafile_sync_error_id_to_str():
        pass
    sync_error_id_to_str = seafile_sync_error_id_to_str

    @searpc_func("int", ["int"])
    def seafile_del_file_sync_error_by_id():
        pass
    del_file_sync_error_by_id = seafile_del_file_sync_error_by_id

    @searpc_func("int", ["string"])
    def seafile_calc_dir_size(path):
        pass
    calc_dir_size = seafile_calc_dir_size

    @searpc_func("int64", [])
    def seafile_get_total_block_size():
        pass
    get_total_block_size = seafile_get_total_block_size;

    @searpc_func("string", ["string"])
    def seafile_get_config(key):
        pass
    get_config = seafile_get_config

    @searpc_func("int", ["string", "string"])
    def seafile_set_config(key, value):
        pass
    set_config = seafile_set_config

    @searpc_func("int", ["string"])
    def seafile_get_config_int(key):
        pass
    get_config_int = seafile_get_config_int

    @searpc_func("int", ["string", "int"])
    def seafile_set_config_int(key, value):
        pass
    set_config_int = seafile_set_config_int

    @searpc_func("int", ["int"])
    def seafile_set_upload_rate_limit(limit):
        pass
    set_upload_rate_limit = seafile_set_upload_rate_limit

    @searpc_func("int", ["int"])
    def seafile_set_download_rate_limit(limit):
        pass
    set_download_rate_limit = seafile_set_download_rate_limit

    ### repo
    @searpc_func("objlist", ["int", "int"])
    def seafile_get_repo_list():
        pass
    get_repo_list = seafile_get_repo_list

    @searpc_func("object", ["string"])
    def seafile_get_repo():
        pass
    get_repo = seafile_get_repo

    @searpc_func("string", ["string", "string", "string", "string", "string", "int"])
    def seafile_create_repo(name, desc, passwd, base, relay_id, keep_history):
        pass
    create_repo = seafile_create_repo

    @searpc_func("int", ["string"])
    def seafile_destroy_repo(repo_id):
        pass
    remove_repo = seafile_destroy_repo

    @searpc_func("objlist", ["string", "string", "string", "int"])
    def seafile_diff():
        pass
    get_diff = seafile_diff

    @searpc_func("object", ["string", "int", "string"])
    def seafile_get_commit(repo_id, version, commit_id):
        pass
    get_commit = seafile_get_commit

    @searpc_func("objlist", ["string", "int", "int"])
    def seafile_get_commit_list():
        pass
    get_commit_list = seafile_get_commit_list

    @searpc_func("objlist", ["string"])
    def seafile_branch_gets(repo_id):
        pass
    branch_gets = seafile_branch_gets

    @searpc_func("int", ["string", "string"])
    def seafile_branch_add(repo_id, branch):
        pass
    branch_add = seafile_branch_add

    ##### clone related
    @searpc_func("string", ["string", "string"])
    def gen_default_worktree(worktree_parent, repo_name):
        pass

    @searpc_func("string", ["string", "int", "string", "string", "string", "string", "string", "string", "int", "string"])
    def seafile_clone(repo_id, repo_version, repo_name, worktree, token, password, magic, email, random_key, enc_version, more_info):
        pass
    clone = seafile_clone

    @searpc_func("string", ["string", "int", "string", "string", "string", "string", "string", "string", "int", "string"])
    def seafile_download(repo_id, repo_version, repo_name, wt_parent, token, password, magic, email, random_key, enc_version, more_info):
        pass
    download = seafile_download

    @searpc_func("int", ["string"])
    def seafile_cancel_clone_task(repo_id):
        pass
    cancel_clone_task = seafile_cancel_clone_task

    @searpc_func("objlist", [])
    def seafile_get_clone_tasks():
        pass
    get_clone_tasks = seafile_get_clone_tasks

    @searpc_func("object", ["string"])
    def seafile_find_transfer_task(repo_id):
        pass
    find_transfer_task = seafile_find_transfer_task

    ### sync
    @searpc_func("int", ["string", "string"])
    def seafile_sync(repo_id, peer_id):
        pass
    sync = seafile_sync

    @searpc_func("object", ["string"])
    def seafile_get_repo_sync_task():
        pass
    get_repo_sync_task = seafile_get_repo_sync_task

    @searpc_func("int", [])
    def seafile_is_auto_sync_enabled():
        pass
    is_auto_sync_enabled = seafile_is_auto_sync_enabled

    @searpc_func("objlist", ["int", "int"])
    def seafile_get_file_sync_errors():
        pass
    get_file_sync_errors = seafile_get_file_sync_errors

    ###### Property Management #########

    @searpc_func("int", ["string", "string"])
    def seafile_set_repo_passwd(repo_id, passwd):
        pass
    set_repo_passwd = seafile_set_repo_passwd

    @searpc_func("int", ["string", "string", "string"])
    def seafile_set_repo_property(repo_id, key, value):
        pass
    set_repo_property = seafile_set_repo_property

    @searpc_func("string", ["string", "string"])
    def seafile_get_repo_property(repo_id, key):
        pass
    get_repo_property = seafile_get_repo_property

    @searpc_func("string", ["string"])
    def seafile_get_repo_relay_address(repo_id):
        pass
    get_repo_relay_address = seafile_get_repo_relay_address

    @searpc_func("string", ["string"])
    def seafile_get_repo_relay_port(repo_id):
        pass
    get_repo_relay_port = seafile_get_repo_relay_port

    @searpc_func("int", ["string", "string", "string"])
    def seafile_update_repo_relay_info(repo_id, addr, port):
        pass
    update_repo_relay_info = seafile_update_repo_relay_info

    @searpc_func("int", ["string", "string"])
    def seafile_set_repo_token(repo_id, token):
        pass
    set_repo_token = seafile_set_repo_token

    @searpc_func("string", ["string"])
    def seafile_get_repo_token(repo_id):
        pass
    get_repo_token = seafile_get_repo_token

    @searpc_func("object", ["int", "string", "string"])
    def seafile_generate_magic_and_random_key(enc_version, repo_id, password):
        pass
    generate_magic_and_random_key = seafile_generate_magic_and_random_key

    @searpc_func("int", [])
    def seafile_shutdown():
        pass
    shutdown = seafile_shutdown

    @searpc_func("int", ["string", "int"])
    def seafile_add_del_confirmation(key, value):
        pass
    add_del_confirmation = seafile_add_del_confirmation



"""
Define RPC functions needed to generate
"""

# [ <ret-type>, [<arg_types>] ]
func_table = [
    [ "int", [] ],
    [ "int", ["int"] ],
    [ "int", ["int", "int"] ],
    [ "int", ["int", "string"] ],
    [ "int", ["int", "string", "int"] ],
    [ "int", ["int", "string", "string"] ],
    [ "int", ["int", "string", "int", "int"] ],    
    [ "int", ["int", "int", "string", "string"] ],
    [ "int", ["string"] ],
    [ "int", ["string", "int"] ],
    [ "int", ["string", "int", "int"] ],
    [ "int", ["string", "int", "string"] ],
    [ "int", ["string", "int", "string", "string"] ],
    [ "int", ["string", "int", "int", "string", "string"] ],
    [ "int", ["string", "string"] ],
    [ "int", ["string", "string", "string"] ],
    [ "int", ["string", "string", "int", "int"] ],
    [ "int", ["string", "string", "string", "int"] ],
    [ "int", ["string", "string", "string", "int", "string"] ],
    [ "int", ["string", "string", "string", "string"] ],
    [ "int", ["string", "string", "string", "string", "string"] ],
    [ "int", ["string", "string", "string", "string", "string", "string"] ],
    [ "int", ["string", "string", "string", "int", "string", "string"] ],
    [ "int", ["string", "string", "string", "string", "string", "string", "string"] ],
    [ "int", ["string", "int64"]],
    [ "int", ["int", "int64"]],
    [ "int", ["int", "string", "int64"]],
    [ "int64", [] ],
    [ "int64", ["string"] ],
    [ "int64", ["int"]],
    [ "int64", ["int", "string"]],
    [ "int64", ["string", "int", "string"] ],
    [ "string", [] ],
    [ "string", ["int"] ],
    [ "string", ["int", "int"] ],
    [ "string", ["int", "string"] ],
    [ "string", ["int", "int", "string"] ],
    [ "string", ["string"] ],
    [ "string", ["string", "int"] ],
    [ "string", ["string", "int", "int"] ],
    [ "string", ["string", "string"] ],
    [ "string", ["string", "string", "int"] ],
    [ "string", ["string", "string", "int", "int"] ],
    [ "string", ["string", "string", "string"] ],
    [ "string", ["string", "string", "string", "string"] ],
    [ "string", ["string", "string", "string", "string", "int"] ],
    [ "string", ["string", "string", "string", "string", "string"] ],
    [ "string", ["string", "string", "string", "string", "string", "int"] ],
    [ "string", ["string", "string", "string", "string", "string", "string", "int"] ],
    [ "string", ["string", "string", "string", "string", "string", "string", "int", "int"] ],
    [ "string", ["string", "string", "string", "string", "string", "string"] ],
    [ "string", ["string", "string", "string", "string", "string", "string", "int64"] ],
    [ "string", ["string", "string", "string", "string", "string", "string", "int64", "int"] ],
    [ "string", ["string", "string", "string", "string", "string", "string", "string"] ],
    [ "string", ["string", "string", "string", "string", "string", "string", "string", "int64"] ],
    [ "string", ["string", "string", "string", "string", "string", "string", "string", "string", "string"] ],
    [ "string", ["string", "int", "string", "string", "string", "string", "string", "string", "string", "int", "string"] ],
    [ "string", ["string", "int", "string", "int", "int"] ],
    [ "string", ["string", "int", "string", "string", "string"] ],
    [ "objlist", [] ],
    [ "objlist", ["int"] ],
    [ "objlist", ["int", "int"] ],
    [ "objlist", ["int", "string"] ],
    [ "objlist", ["int", "int", "int"] ],
    [ "objlist", ["string"] ],        
    [ "objlist", ["string", "int"] ],
    [ "objlist", ["string", "int", "int"] ],
    [ "objlist", ["string", "int", "string"] ],
    [ "objlist", ["string", "string"] ],        
    [ "objlist", ["string", "string", "string"] ],
    [ "objlist", ["string", "string", "int"] ],
    [ "objlist", ["string", "string", "string", "int"] ],
    [ "objlist", ["string", "string", "int", "int"] ],
    [ "objlist", ["string", "string", "int", "int", "int"] ],
    [ "objlist", ["int", "string", "string", "int", "int"] ],
    [ "objlist", ["string", "int", "string", "string", "string"] ],
    [ "objlist", ["string", "int", "string", "int", "int"] ],
    [ "objlist", ["string", "int", "string", "string", "int"] ],
    [ "objlist", ["string", "string", "string", "string", "int", "int"] ],
    [ "object", ["int"] ],
    [ "object", ["string"] ],
    [ "object", ["string", "string"] ],
    [ "object", ["string", "string", "string"] ],
    [ "object", ["string", "int", "string"] ],
    [ "object", ["int", "string", "string"] ],
    [ "object", ["int", "string", "string", "string", "string"] ],
    [ "object", ["string", "string", "string", "string", "string", "string", "string", "int", "int"] ],
    [ "object", ["string", "string", "string", "string", "string", "string", "int", "string", "int", "int"] ],
    [ "json", [] ],
]
